package org.gel.mauve;

import java.awt.event.*;
import java.applet.Applet;
import java.net.*;
import java.util.*;
import java.io.*;
import java.awt.*;
import javax.swing.*;

public class AnimatedRulerPanel extends RulerPanel implements Runnable
{
	int fps = 20;	//**< Frames per second for animation */
    int frame = 0;
    int delay;
	
    Dimension offDimension;
    Image offImage;
    Graphics offGraphics;
	
	int move_subticks = 0;
	int zoom_subticks = 0;

	int tick_offset = 0;
	int tick_frames = 50;	/**< Number of frames to scroll a tick */
	
	public AnimatedRulerPanel( int a, int b, int c, int d ){
		super( a, b, c, d );
	}
	
	public boolean isDoubleBuffered() {
		return true;
	}
	
	synchronized public void Zoom( int sub_ticks ){
		zoom_subticks += sub_ticks;
		notify();
	}

	synchronized public void Move( int sub_ticks ){
		move_subticks += sub_ticks;
		notify();
	}
	synchronized public void StopAnimation(){
		if( zoom_subticks < 0 )
			zoom_subticks = -1;
		else if ( zoom_subticks > 0 )
			zoom_subticks = 1;

		if( move_subticks < 0 )
			move_subticks = -1;
		else if ( move_subticks > 0 )
			move_subticks = 1;

	}
	synchronized protected int GetZoomSubticks(){
		return zoom_subticks;
	}

	synchronized protected int GetMoveSubticks(){
		return move_subticks;
	}
	
	synchronized protected void WaitForAction(){
		try{
			wait(0);
		}catch( InterruptedException ie ) {
			ie.printStackTrace();
		}
	}
	
/*	synchronized protected int GetFirstLabel(){
		return FirstViewable;
	}

	synchronized protected int GetLastLabel(){
		return LastViewable;
	}
*/

    /**
     * This method is called by the thread that was created in
     * the start method. It does the main animation.
     */
	public void run() {
		delay = (fps > 0) ? (1000 / fps) : 100;

		// create the offscreen graphics context
		Dimension d = getSize();
	    offDimension = d;
	    offImage = createImage(d.width, d.height);
	    offGraphics = offImage.getGraphics();
		
		// have the superclass paint the initial ruler
		// paint the initial ruler
		paintFrame( offGraphics );
		repaint();
		
		// Remember the starting time
		while( true ){
			WaitForAction();
			long tm = System.currentTimeMillis();
			while ( GetZoomSubticks() != 0 || GetMoveSubticks() != 0 ) {
			    // Display the next frame of animation.
			    repaint();

			    // Delay depending on how far we are behind.
			    try {
					tm += delay;
					Thread.sleep(Math.max(0, tm - System.currentTimeMillis()));
			    } catch (InterruptedException e) {
					break;
			    }

			    // Advance the frame
			    frame++;
			    if( frame == tick_frames ){
			    	if( zoom_subticks > 0 )
			    		zoom_subticks--;
			    	else if( zoom_subticks < 0 )
			    		zoom_subticks++;
			    	if( move_subticks > 0 )
			    		move_subticks--;
			    	else if( move_subticks < 0 )
			    		move_subticks++;
			    	frame = 0;
			    	
//			    	super.paint( offGraphics );
			    }
			}
		}
    }

    /**
     * Update a frame of animation.
     */
    public void update(Graphics g) {
		Dimension d = getSize();

		// Update the offscreen graphics context if necessary
		if ((offGraphics == null)
		 || (d.width != offDimension.width)
		 || (d.height != offDimension.height)) {
		    offDimension = d;
		    offImage = createImage(d.width, d.height);
		    offGraphics = offImage.getGraphics();
			
			// have the superclass paint the initial ruler
//		    super.paint(offGraphics);
		}
		
		// erase the previous image...
		offGraphics.setColor(getBackground());
		offGraphics.fillRect(0, 0, d.width, d.height);
		offGraphics.setColor(Color.black);

		double tickerwidth = getSize().width / (double)ticks;
		double tickmove = tickerwidth / tick_frames;
		tick_offset = - (int)(tickmove)*frame;
		paintFrame( offGraphics );
		
		// Paint the image onto the screen
		g.drawImage(offImage, 0, 0, null);
    }

    /**
     * Paint the previous frame (if any).
     */
    public void paint(Graphics g) {
//    	System.out.println(" painting ");
		if (offImage != null) {
		    g.drawImage(offImage, 0, 0, null);
		}
    }

    /**
     * Paint a frame of animation.
     */
    public void paintFrame (Graphics g) {
		double tickerval = (double)(lastViewable - firstViewable) / (double)ticks;
		double tickerwidth = getSize().width / (double)ticks;
		double subtickerwidth = tickerwidth / (double)subticks;
		int midpoint = getSize().width / 2;

		// draw the base line
		g.setColor( lineColor );
		g.drawLine( 0, topOffset, getSize().width, topOffset );

		//draw the left half, then draw the right half.
		int tickBase = 0 + tick_offset;
		int drawTicks = ticks / 2;
		if( ticks % 2 == 1 )
			drawTicks++;
		
		for ( int halfI = 0; halfI < 2; halfI++ ){
			if( halfI == 1 )
				tickBase = midpoint - tick_offset;
			// draw the ticks and subticks
			int subtickX;
			for ( int i = 0; i < drawTicks; i++ ){
				int tickX = (int)(tickerwidth * i + tickBase);
				if( tickX >= 0 && tickX <= getSize().width )
				    g.drawLine( tickX, topOffset , tickX, topOffset + tickHeight);

				for ( int subI = 0; subI < subticks; subI++ ){
					subtickX = (int)(subtickerwidth * subI + tickX);
					if( subtickX >= 0  && subtickX <= getSize().width )
					    g.drawLine( subtickX, topOffset , subtickX, topOffset + subtickHeight);
				}
			}
			
			// draw the tick labels
		    String i_string;
			g.setColor( textColor );
			for ( int i = 0; i < drawTicks; i++ ){
				int tickX = (int)(tickerwidth * i + tickBase);
				if( tickX >= 0 && tickX <= getSize().width && (i + 1 < drawTicks || halfI == 0) ){
				    i_string = new Integer((int)((i * tickerval) + firstViewable)).toString();
				    g.drawString(i_string, tickX + 2, labelOffset); 
				}
	 		}
	 	}
    }

	public void paintComponent( Graphics g ) {
		paint( g );
	}

}
