package org.gel.mauve;
import java.util.*;

/**
 * Key specifies the interface allowed for intervals that can be stored in an
 * interval sequence tree.
 */
interface Key {
	public long getLength();
	public long getSeqLength();
	public void cropStart( long size );
	public void cropEnd( long size );
	public Object clone();
	public void print();
}

class IstIterator implements Iterator {
	IstNode node;
	IntervalSequenceTree ist;

	IstIterator( IntervalSequenceTree ist, IstNode node ) {
		this.ist = ist;
		this.node = node;
	}

//	Returns true if the iteration has more elements.
	public boolean	hasNext() {
		return ist.increment( node ) != null;
	}

//	Returns the next element in the iteration.	
	public Object next() {
		node = ist.increment( node );
		return node.key;
	}

//	Removes from the underlying collection the last element returned by the iterator (optional operation).
	public void remove() {
	
	}
}

class IstNode {
	IstNode parent;
	IstNode left;
	IstNode right;
	long subtree_size;
	long length;
	Key key;
};


class GapKey implements Key{
	long length;
	GapKey( long length ) {
		this.length = length;
	}
	
	public long getLength() {
		return length;
	}
	public long getSeqLength() {
		return 0;
	}
	
	public void cropStart( long size ){
		length -= size;
		if( length < 0 )
			throw new ArrayIndexOutOfBoundsException();
	}

	public void cropEnd( long size ){
		length -= size;
		if( length < 0 )
			throw new ArrayIndexOutOfBoundsException();
	}
	public Object clone(){
		GapKey gk = new GapKey( length );
		return gk;
	}
	public void print(){
		String info = "GapKey length: " + length;
		System.err.println( info );
	}
}

class StringKey implements Key{
	String seq;
	StringKey( String seq ) {
		this.seq = seq;
	}
	
	public long getLength() {
		return (long)seq.length();
	}
	public long getSeqLength() {
		return (long)seq.length();
	}
	public void cropStart( long size ){
		seq = seq.substring( (int)size );
	}

	public void cropEnd( long size ){
		seq = seq.substring( 0, (int)(getLength() - size) );
	}
	public Object clone(){
		StringKey sk = new StringKey( seq );
		return sk;
	}
	public void print() {
		String info = "StringKey length: " + seq.length();
		System.err.println( info );
	}
}

class FileKey implements Key{
	XMFAAlignment xmfa;
	long lend;
	long length;
	long f_offset;
	long f_length;

	FileKey( XMFAAlignment xmfa, long lend, long length, long f_offset, long f_length ) {
		this.xmfa = xmfa;
		this.lend = lend;
		this.length = length;
		this.f_offset = f_offset;
		this.f_length = f_length;
	}
	
	public long getLength() {
		return length;
	}
	public long getSeqLength() {
		return length;
	}
	
	public void cropStart( long size ){
		length -= size;
		lend += size;
		if( length < 0 )
			throw new ArrayIndexOutOfBoundsException();
	}

	public void cropEnd( long size ){
		length -= size;
		if( length < 0 )
			throw new ArrayIndexOutOfBoundsException();
	}
	public Object clone(){
		FileKey fk = new FileKey( xmfa, lend, length, f_offset, f_length );
		return fk;
	}
	public void print() {
		String info = "FileKey seq_offset: " + lend + " seq_len: " + length + " bfo: " + f_offset + " len: " + f_length;
		System.err.println( info );
	}
}


class IntervalSequenceTree {
// SortedMap methods
	public Comparator comparator() {
		return null;
	} 
	public Object firstKey(){
		return null;
	}
	public IntervalSequenceTree headMap( Object toKey ){
		return null;
	}
	public Object lastKey(){
		return null;
	}
	public IntervalSequenceTree subMap( Long fromKey, Long toKey ){
		return null;
	}
	public IntervalSequenceTree tailMap( Long fromKey ){
		return null;
	}

// Map methods
//          Removes all mappings from this map (optional operation).	
	public void clear(){
		root = null;
		leftmost = null;
		rightmost = null;
	}

	//          Returns true if this map contains a mapping for the specified key.	
	public boolean containsKey( Long key ){
		return true;
	}
	//          Returns true if this map maps one or more keys to the specified value.	
	public boolean containsValue(Object value){
		return false;
	}
	//          Returns a set view of the mappings contained in this map.	
	public Set entrySet(){
		return null;
	}
	//          Compares the specified object with this map for equality.	
	public boolean equals(Object o){
		return false;
	}
	//          Returns the value to which this map maps the specified key.	
	public Object get(Long key){
		return null;
	}
	//          Returns the hash code value for this map.	
	public int hashCode() {
		return 0;
	}
//	Returns true if this map contains no key-value mappings.	
	public boolean	isEmpty() {
		return root == null;
	}

//	Returns a set view of the keys contained in this map.	
	public Set	keySet() {
		return null;
	}
//	Associates the specified value with the specified key in this map (optional operation).	
	public Object put(Object key, Object value) {
		return null;
	}
//	Copies all of the mappings from the specified map to this map (optional operation).	
	public void putAll(Map t) {}
//	Removes the mapping for this key from this map if it is present (optional operation).	
	public Object remove(Object key) {
		return null;
	}
//	Returns the number of key-value mappings in this map.	
	public long size() {
		return root == null ? 0 : root.length;
	}
//	Returns a collection view of the values contained in this map.
	public Collection values() {
		return null;
	}


// interval sequence specific:
	/**
	 * Returns the total length of intervals contained in this interval sequence
	 */
//	long length();
	long nodeCount(){
		return root == null ? 0 : root.subtree_size;
	}
//	long countNodes( IstNode* x = NULL );

	IstNode root;		/**< Root of the tree */
	IstNode leftmost;	/**< Left most tree node, for begin() method */
	IstNode rightmost;	/**< Right most tree node, for end() method */



	void checkTree( IstNode cur_node ){
		if( cur_node != null ){
			if( cur_node.parent == cur_node )
				System.err.println( "freakout" );
			checkTree( cur_node.left );
			checkTree( cur_node.right );
		}
	}

	public Iterator insert( Key val, long point )
	{
		long[] iv_offset = new long[1];
		iv_offset[0] = point;
		IstNode ins_node = recursiveFind( iv_offset, root );
		IstNode new_node = new IstNode();
		new_node.key = (Key)val.clone();
		new_node.length = val.getLength();
		new_node.subtree_size = 0;
		if( ins_node == null ){
			// end insert
			rightmost = new_node;
			if( root == null ){
				root = new_node;
				leftmost = new_node;
				return new IstIterator( this, new_node );
			}
			// find the shallowest right insertion point
			ins_node = decrement( null );
			// make a new parent node
			IstNode new_parent = new IstNode();
			new_parent.left = ins_node;
			new_parent.right = new_node;
			new_parent.parent = ins_node.parent;
			if( new_parent.parent == null )
				root = new_parent;
			else
				new_parent.parent.right = new_parent;
			ins_node.parent = new_parent;
			new_parent.length = ins_node.length;

			// update lengths and subtree_sizes along the path to the root
	//		checkTree( root );
	//		propogateChanges( new_node, 0, 0 );
	//		propogateChanges( ins_node, 0, 0 );
			propogateChanges( new_parent, new_node.length, 2 );
			return new IstIterator( this, new_node );
		}

		// iv_offset is the distance into the node that the leaf should be split
		// 0 is a special case (left insert)
		if( iv_offset[0] == 0 ){
			IstNode new_parent = new IstNode();
			new_parent.left = new_node;
			new_parent.right = ins_node;
			new_parent.parent = ins_node.parent;
			if( new_parent.parent.right == ins_node )
				new_parent.parent.right = new_parent;
			else
				new_parent.parent.left = new_parent;
			new_parent.length = ins_node.length;

			ins_node.parent = new_parent;
			new_node.parent = new_parent;

			if( point == 0 )
				leftmost = new_node;
			// update lengths and subtree_sizes along the path to the root
	//		checkTree( root );
	//		propogateChanges( new_node, 0, 0 );
	//		propogateChanges( ins_node, 0, 0 );
			propogateChanges( new_parent, new_node.length, 2 );
		}else{
			// need to split a leaf node
			IstNode new_gp = new IstNode();
			IstNode new_parent = new IstNode();
			new_gp.parent = ins_node.parent;
			new_gp.right = new_parent;
			new_gp.left = new IstNode();
			new_gp.left.key = (Key)ins_node.key.clone();
			new_gp.left.key.cropEnd( ins_node.length - iv_offset[0] );
			new_gp.left.length = new_gp.left.key.getLength();
			new_gp.left.parent = new_gp;
			
			ins_node.key.cropStart( iv_offset[0] );
			ins_node.length = ins_node.key.getLength();
			ins_node.parent = new_parent;
			new_node.parent = new_parent;
			new_parent.left = new_node;
			new_parent.right = ins_node;
			new_parent.parent = new_gp;
			new_parent.length = new_node.length + ins_node.length;
			new_parent.subtree_size = 2;

			new_gp.length = ins_node.length + new_gp.left.length;
			new_gp.subtree_size = 1;
			if( new_gp.parent == null ){
				root = new_gp;
				leftmost = new_gp.left;
				rightmost = ins_node;
			}else if( new_gp.parent.right == ins_node )
				new_gp.parent.right = new_gp;
			else
				new_gp.parent.left = new_gp;

			// update lengths and subtree_sizes along the path to the root
	//		checkTree( root );
	//		propogateChanges( new_node, 0, 0 );
	//		propogateChanges( ins_node, 0, 0 );
	//		propogateChanges( new_gp->left, 0, 0 );
	//		propogateChanges( new_parent, 0, 0 );
			new_gp.subtree_size = -1;
			propogateChanges( new_gp, new_node.length, 4 );
		}
		return new IstIterator( this, new_node );
	}

	public long erase( long point, long length )
	{
		long[] iv_offset = new long[1];
		iv_offset[0] = point;
		
		IstNode ins_node = recursiveFind( iv_offset, root );

		// iv_offset is the distance into the node that the leaf should be split
		// 0 is a special case (left delete)
		long deleted_nodes = 0;
		while( length > 0 ){
			if( ins_node == null ){
				// end delete?  that's illegal
				return deleted_nodes;
			}
			if( iv_offset[0] == 0 ){
				if( length >= ins_node.length ){
					// delete the whole thing
					length -= ins_node.length;
					if( ins_node.parent == null ){
						// deleting the root
						root = null;
						leftmost = null;
						rightmost = null;
						return deleted_nodes + 1;
					}

					IstNode other_child = null, del_node = null;
					if( ins_node.parent.left == ins_node ){
						other_child = ins_node.parent.right;
					}else if( ins_node.parent.right == ins_node ){
						other_child = ins_node.parent.left;
					}
					del_node = ins_node;
					ins_node = increment( ins_node );

					// update tree structure
					IstNode tmp_parent = other_child.parent;
					IstNode tmp_gp = tmp_parent.parent;
// java: do something about the = operator!!
					tmp_parent.key = other_child.key;
					tmp_parent.left = other_child.left;
					tmp_parent.right = other_child.right;
					tmp_parent.parent = other_child.parent;
					tmp_parent.subtree_size = other_child.subtree_size;
					tmp_parent.length = other_child.length;

					tmp_parent.parent = tmp_gp;
					if( tmp_parent.left != null )
						tmp_parent.left.parent = tmp_parent;
					if( tmp_parent.right != null )
						tmp_parent.right.parent = tmp_parent;
					if( ins_node == other_child )
						ins_node = tmp_parent;
					
					// propogate deletion length thru root
					tmp_parent = tmp_parent.parent;
	//				checkTree( root );
					propogateChanges( tmp_parent, -del_node.length, -2 );

					++deleted_nodes;
				}else{
					// crop from start
					ins_node.key.cropStart( length );
	//				checkTree( root );
					propogateChanges( ins_node, -length, 0 );
					return deleted_nodes;
				}
			}else if( length >= ins_node.length - iv_offset[0] ){
				// crop from end
				ins_node.key.cropEnd( ins_node.length - iv_offset[0] );
				length -= ins_node.length - iv_offset[0];
	//			checkTree( root );
				propogateChanges( ins_node, -(ins_node.length - iv_offset[0]), 0 );
				ins_node = increment( ins_node );
				iv_offset[0] = 0;
			}else{
				// delete from middle (nastee part)
				IstNode new_parent = new IstNode();
				new_parent.left = ins_node;
				new_parent.length = ins_node.length;
				new_parent.right = new IstNode();
				new_parent.right.key = (Key)ins_node.key.clone();
				new_parent.right.length = ins_node.length - length - iv_offset[0];
				new_parent.right.key.cropStart( length + iv_offset[0] );
				new_parent.left.key.cropEnd( ins_node.length - iv_offset[0] );
				new_parent.left.length = iv_offset[0];
				new_parent.parent = ins_node.parent;
				if( new_parent.parent == null ){
					root = new_parent;
					rightmost = new_parent.right;
				}else if( new_parent.parent.left == ins_node )
					new_parent.parent.left = new_parent;
				else if( new_parent.parent.right == ins_node )
					new_parent.parent.right = new_parent;
				
				ins_node.parent = new_parent;
				new_parent.right.parent = new_parent;
	//			checkTree( root );
	//			propogateChanges( ins_node, 0, 0 );
	//			propogateChanges( new_parent.right, 0, 0 );
				propogateChanges( new_parent, -length, 2 );
				return deleted_nodes;
			}
		}
		return deleted_nodes;
	}

	/**
	 * propogates changes to node values up a tree
	 */
	void propogateChanges( IstNode cur_node, long length_diff, long subtree_diff )
	{
		while( cur_node != null ){
			if( cur_node.parent == cur_node )
				System.err.println( "when I say oh, you say shit!  OH!" );
			cur_node.length += length_diff;
			cur_node.subtree_size += subtree_diff;
			cur_node = cur_node.parent;
		}
	}

	protected IstNode recursiveFind( long[] point, IstNode node ) {
		if( node == null )
			return null;

		// return this node if it's a leaf
		if( node.key != null )
			return node;
		// look for the next node to recurse to
		if( point[0] < node.length ){
			if( node.left != null ){
				if( point[0] < node.left.length )
					return recursiveFind( point, node.left );
				point[0] -= node.left.length;
			}
			return recursiveFind( point, node.right );
		}
		point[0] -= node.length;
		// out of range
		return null;
	}

	IstNode increment( IstNode x) {
		// find the least-ancestor with another child
		// and set x to that child
		while( x.parent != null ){
			if( x == x.parent.left &&
				x.parent.right != null ){
				x = x.parent.right;
				break;
			}else
				x = x.parent;
		}
		// if there were no other children to the right then we're at the end
		if( x.parent == null ){
			x = null;
			return x;
		}

		// find the left-most leaf node below x
		while( x.key == null ){
			if( x.left != null )
				x = x.left;
			else if( x.right != null )
				x = x.right;
		}
		return x;
	}

	IstNode decrement( IstNode x) {
		if( x != null ){
			// find the least-ancestor with another child to the left
			// and set x to that child
			while( x.parent != null ){
				if( x == x.parent.right &&
					x.parent.left != null){
					x = x.parent.left;
					break;
				}else
					x = x.parent;
			}
			// if there was no other children to the left then we're at the end
			// raise hell! (cause an access violation)
			if( x.parent == null )
				x = null;
		}else{
			x = root;
		}
		
		// find the right-most leaf node below x
		while( x.key == null ){
			if( x.right != null )
				x = x.right;
			else if( x.left != null )
				x = x.left;
		}
		return x;
	}
}	
