package org.gel.mauve;


import java.awt.event.*;
import java.applet.Applet;
import java.net.*;
import java.util.*;
import java.io.*;
import java.awt.*;

class Match {
	static final int NO_MATCH = 0;
	
	public Match( int sequenceCount ) {
		starts = new long[ sequenceCount ];
		lengths = new long[ sequenceCount ];
		reverse = new boolean[ sequenceCount ];
		lcb = 0;
	}
	
	public Match( Match m ) {
		starts = new long[ m.starts.length ];
		lengths = new long[ m.lengths.length ];
		reverse = new boolean[ m.reverse.length ];
		System.arraycopy( m.starts, 0, starts, 0, starts.length );
		System.arraycopy( m.lengths, 0, lengths, 0, lengths.length );
		System.arraycopy( m.reverse, 0, reverse, 0, reverse.length );

		color = m.color;
		lcb = m.lcb;
	}
	
	/**< Compute and return the generalized offset */
	public long offset(){
		int seqI = 0;
		long ref;
		long g_offset = 0;
		for( ; seqI < starts.length; seqI++ ){
			if( starts[ seqI ] != NO_MATCH )
				break;
		}
		ref = starts[ seqI ];
		for( ; seqI < starts.length; seqI++ ){
			long cur_start = starts[seqI];
			if( reverse[ seqI ] )
				cur_start = -cur_start;
			g_offset += ref - cur_start;
		}

		return g_offset;
	}
	
	public String toString(){
		String rval = new String();
		for( int seqI = 0; seqI < starts.length; seqI++ ){
			rval += "<";
			if( starts[seqI] != NO_MATCH ){
				if( reverse[seqI] )
					rval += "-";
				rval += starts[seqI];
				rval += ",";
				rval += lengths[seqI];
			}
			rval += "> ";
		}
		return rval;
	}

	public long multiplicityType(){
		// determine the match's multiplicity type.
		long mult_type = 0;
		for( int seqI = 0; seqI < starts.length; seqI++ ){
			mult_type <<= 1;
			if( starts[ seqI ] != Match.NO_MATCH )
				mult_type |= 1;
		}
		return mult_type;
	}
	
	public void reorder( int[] new_order ){
		long[] tmp = null;
		int seqI = 0;

		// reorder starts
		tmp = new long[ new_order.length ];
		for( seqI = 0; seqI < new_order.length; seqI++ )
			tmp[ seqI ] = starts[ new_order[ seqI ] ];
		starts = tmp;

		// reorder lengths
		tmp = new long[ new_order.length ];
		for( seqI = 0; seqI < new_order.length; seqI++ )
			tmp[ seqI ] = lengths[ new_order[ seqI ] ];
		lengths = tmp;
		
		// reorder reverse
		boolean[] rev_tmp = new boolean[ new_order.length ];
		for( seqI = 0; seqI < new_order.length; seqI++ )
			rev_tmp[ seqI ] = reverse[ new_order[ seqI ] ];
		reverse = rev_tmp;
		
	}

	long[] starts;	//*< The start coordinate of this match in each sequence */
	long[] lengths;	//*< The lengths of this match in each sequence */
	boolean[] reverse;	//*< The direction of each match.  false is forward, true is reverse */
	Color color;	//*< The color code of this match */
	boolean highlighted = false;	//*< true if the match is highlighted on screen */
	int lcb;	//*< The lcb which this match belongs to */
}
