/** 
 * Mauve.java
 *
 * Title:			Mauve
 * Description:		Viewer for multiple genome alignments and annotation
 * @author			koadman
 * @version			
 */

package org.gel.mauve;
import javax.swing.*;
import java.util.*;
import java.net.*;
import java.io.*;

public class Mauve extends javax.swing.JApplet {
	// applet functions
	boolean isStandalone = false;

	// Retrieve the value of an applet parameter
	public String getParameter(String key, String def) {
		return isStandalone ? System.getProperty(key, def) :
			(getParameter(key) != null ? getParameter(key) : def);
	}

	// Get info on the applet parameters
	public String[][] getParameterInfo() {
		String info[][] =
		{
			{"alignment_name", "String", ""}
		};

		return info;
	}

	// Get applet information
	public String getAppletInfo() {
		return "Applet Information";
	}
	
	/**
	 * Checks the Mauve web server for a newer version of this software
	 */
	public void checkForUpdates() {
		try{
			URL hurl = new URL( "http://gel.ahabs.wisc.edu/mauve/downloads/latest" );
			BufferedReader latest_in = new BufferedReader(new InputStreamReader(hurl.openStream()));
			long latest_version = Long.parseLong( latest_in.readLine() );
			
			FileInputStream local_dat = new FileInputStream("version");
			BufferedReader local_in = new BufferedReader( new InputStreamReader( local_dat ) );
			long local_version = Long.parseLong( local_in.readLine() );
			if( local_version < latest_version ){
				// check the OS type since linux can't do an auto-update
				String os_type = System.getProperty( "os.name" );
				if( !os_type.startsWith( "Windows" ) ){
					JOptionPane.showConfirmDialog(null, "An updated version of Mauve is available.\nSee the Mauve web site at http://gel.ahabs.wisc.edu/mauve for more details...", "Updated Mauve available", JOptionPane.YES_NO_OPTION);
					return;
				}

				int dl_val = JOptionPane.showConfirmDialog(null, "An updated version of Mauve is available.\nWould you like to download and install it?", "Updated Mauve available", JOptionPane.YES_NO_OPTION);
				if( dl_val == 0 ){
					//
					// download the installer to a temp file
					//
					System.out.println( "Downloading installer...\n" );
					String dl_location = "http://gel.ahabs.wisc.edu/mauve/downloads/mauve_installer_";
					dl_location += Long.toString( latest_version ) + ".exe";
					URL dlurl = new URL( dl_location );
					InputStream is = dlurl.openStream();
					File ins_file = File.createTempFile( "mauve_ins", ".exe" );
					FileOutputStream output_file = new FileOutputStream ( ins_file );
					byte[] buf = new byte[ 1024*1024 ];
					int read_chars = is.read( buf );
					while( read_chars >= 0 ){
						output_file.write( buf, 0, read_chars );
						read_chars = is.read( buf );
					}
					// close the output file
					output_file.close();
					//
					// run the installer
					//
					String[] ins_cmd = new String[ 1 ];
					ins_cmd[ 0 ] = ins_file.getPath();
					Process ins_proc = Runtime.getRuntime().exec( ins_cmd );
					// delete the installer
					ins_file.delete();
					// exit this outdated pabbis abbof shabbit
					System.exit(0);
				}
			}
		}catch( Exception e ){
			System.err.println( "Error checking for updates:" );
			e.printStackTrace();
		}
	}

	// Initialize the applet
	public void init() {
		// 
		// First check the JVM version to make sure it's modern enough
		// 
		String jvm_version = System.getProperty("java.version");
//		System.out.println( jvm_version );
		int minor_version = Integer.parseInt( jvm_version.substring( 2, 3 ) );
		if( jvm_version.charAt( 0 ) == '1' && minor_version < 2 ){
			System.err.println( "Sorry, Mauve requires at least Java version 1.4 to operate correctly" );
		}
		if( jvm_version.charAt( 0 ) == '1' && minor_version == 2 ){
			// bomb out on the user
			JOptionPane.showMessageDialog(null, "Mauve is running in an outdated Java environment (version " + jvm_version + ").\nMauve requires at least Java version 1.4 to function correctly.", "Error", JOptionPane.ERROR_MESSAGE);
			return;
		}
		if( jvm_version.charAt( 0 ) == '1' && minor_version == 3 ){
			// show a warning message
			JOptionPane.showMessageDialog(null, "Mauve is running in an outdated Java environment (version " + jvm_version + ").\nAlthough you may continue to use Mauve, it operates best under Java version 1.4 or later.", "Warning!", JOptionPane.ERROR_MESSAGE);
		}

		// standalone Mauve needs to check for updates
		if( isStandalone ){
			checkForUpdates();
		}
		
		try {
			// For native Look and Feel, uncomment the following code.

			try {
//				UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
			} 
			catch (Exception e) { 
			}
			// pop up the mauve window in here
			frames = new Vector();
			MauveFrame frame = new MauveFrame( this );
			frame.pack();
			frame.initComponents();
			frame.setVisible(true);
			addFrame( frame );
			if( !isApplication() )
				frame.loadURL( getParameter( "alignment_name", "" ) );
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}
	boolean isApplication() {
		return isStandalone;
	}

	public Mauve() {}
	
	private Vector frames;

	void addFrame( MauveFrame frame ){
		frames.addElement( frame );
	}
	
	int getFrameCount() {
		return frames.size();
	}
	
	void removeFrame( MauveFrame frame ){
		for( int frameI = 0; frameI < frames.size(); frameI++ ){
			if( (MauveFrame)(frames.elementAt( frameI )) == frame ){
				frames.remove( frameI );
				break;
			}
		}
		if( frames.size() == 0 )
			System.exit(0);	// no more frames left.  exit.
	}

	// Main entry point
	static public void main(String[] args) {
		Mauve m = new Mauve();
		m.isStandalone = true;
		m.init();
	}
	

    /**
     * Tokenize parameters in @param paramString on tabs, pipes, greater-than, and less-than and put them
     * in a Vector.
     */
    static public Vector parseParams( String paramString ) {
		return parseParams( paramString, "\t|<>" );
    }

    /**
     * Tokenize parameters in @param paramString on the characters in @param delims and put them
     * in a Vector.
     */
    static public Vector parseParams(String paramString, String delims) {
	
		Object ref;
		
		Vector a = new Vector();

		ref = paramString;
		
		if (ref != null) {
		    StringTokenizer st = new StringTokenizer(new String ((String)ref), delims);
		    while (st.hasMoreTokens()) {
				String s = new String (st.nextToken());
				a.addElement(s);
		    }
		}
		return a;
    }
}
