/** 
 * MauveFrame.java
 *
 * Title:			Mauve
 * Description:		Viewer for multiple genome alignments and annotation
 * @author			koadman
 * @version			
 */

package org.gel.mauve;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.io.*;
import java.net.*;
import java.util.*;
import java.beans.*;

public class MauveFrame extends javax.swing.JFrame implements ActionListener, PropertyChangeListener {

// IMPORTANT: Source code between BEGIN/END comment pair will be regenerated
// every time the form is saved. All manual changes will be overwritten.
// BEGIN GENERATED CODE
	// member declarations
	javax.swing.JMenuBar jMenuBar1 = new javax.swing.JMenuBar();
	javax.swing.JMenu jMenuFile = new javax.swing.JMenu();
	javax.swing.JMenuItem jMenuFileOpen = new javax.swing.JMenuItem();
	javax.swing.JMenuItem jMenuFileAlign = new javax.swing.JMenuItem();
	javax.swing.JMenuItem jMenuFilePrint = new javax.swing.JMenuItem();
	javax.swing.JMenuItem jMenuFileClose = new javax.swing.JMenuItem();
	javax.swing.JSeparator jMenuFileSeparator1 = new javax.swing.JSeparator();
	javax.swing.JMenuItem jMenuFileQuit = new javax.swing.JMenuItem();

// END GENERATED CODE

	JFileChooser fc;
	DnDRearrangementPanel rrpanel;
	JToolBar toolbar;
	LCBStatusBar status_bar;
	Mauve mauve;
	AlignFrame alignFrame;

	static Cursor hand_cursor = null;
	static Cursor fist_cursor = null;
	
	static ImageIcon home_button_icon = new ImageIcon( "images/Home16.gif" );
	static ImageIcon left_button_icon = new ImageIcon( "images/Back16.gif" );
	static ImageIcon right_button_icon = new ImageIcon( "images/Forward16.gif" );
	static ImageIcon zoomin_button_icon = new ImageIcon( "images/ZoomIn16.gif" );
	static ImageIcon zoomout_button_icon = new ImageIcon( "images/ZoomOut16.gif" );
	static ImageIcon hand_button_icon = new ImageIcon( "images/Hand16.gif" );
	static ImageIcon dark_hand_button_icon = new ImageIcon( "images/DarkHand16.gif" );

	public MauveFrame( Mauve mauve ) {
		this.mauve = mauve;
		setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
		if( mauve.isApplication() )
			fc = new JFileChooser();
	}

	public void initComponents() throws Exception {
		java.awt.Image mauve_icon = null;
		java.awt.Image hand_icon = null;
		java.awt.Image fist_icon = null;

		if( mauve.isApplication() ){
			mauve_icon = getToolkit().getImage("images/mauve_icon.gif");
			setIconImage( mauve_icon );
			hand_icon = getToolkit().getImage("images/hand_icon.gif");
			fist_icon = getToolkit().getImage("images/fist_icon.gif");
		}else{
//			mauve_icon = getToolkit().getImage(ClassLoader.getSystemResource("mauve_icon.gif"));
		}
		if( hand_icon != null && fist_icon != null ){
			hand_cursor = getToolkit().createCustomCursor( hand_icon, new Point(8, 8), "hand" );
			fist_cursor = getToolkit().createCustomCursor( fist_icon, new Point(8, 8), "fist" );
		}
// IMPORTANT: Source code between BEGIN/END comment pair will be regenerated
// every time the form is saved. All manual changes will be overwritten.
// BEGIN GENERATED CODE
		// the following code sets the frame's initial state
		jMenuBar1.setVisible(true);
		jMenuFile.setToolTipText("Perform file related actions");
		jMenuFile.setVisible(true);
		jMenuFile.setText("File");
		jMenuFile.setMnemonic('F');
		jMenuFileOpen.setToolTipText("Open an existing alignment...");
		jMenuFileOpen.setVisible(true);
		jMenuFileOpen.setText("Open...");
		jMenuFileOpen.setAccelerator(KeyStroke.getKeyStroke( KeyEvent.VK_O, ActionEvent.CTRL_MASK));
		jMenuFileOpen.setMnemonic( 'O' );
		jMenuFileAlign.setToolTipText("Align sequences...");
		jMenuFileAlign.setVisible(true);
		jMenuFileAlign.setText("Align...");
		jMenuFileAlign.setMnemonic( 'A' );
		jMenuFilePrint.setToolTipText("Print the current view...");
		jMenuFilePrint.setVisible(true);
		jMenuFilePrint.setEnabled(false);
		jMenuFilePrint.setText("Print");
		jMenuFilePrint.setAccelerator(KeyStroke.getKeyStroke( KeyEvent.VK_P, ActionEvent.CTRL_MASK));
		jMenuFilePrint.setMnemonic( 'P' );
		jMenuFileClose.setToolTipText("Close this alignment...");
		jMenuFileClose.setVisible(true);
		jMenuFileClose.setText("Close");
		jMenuFileClose.setAccelerator(KeyStroke.getKeyStroke( KeyEvent.VK_W, ActionEvent.CTRL_MASK));
		jMenuFileClose.setMnemonic( 'C' );
		jMenuFileSeparator1.setVisible(true);
		jMenuFileQuit.setToolTipText("Quit this application");
		jMenuFileQuit.setVisible(true);
		jMenuFileQuit.setText("Quit");
		jMenuFileQuit.setAccelerator(KeyStroke.getKeyStroke( KeyEvent.VK_Q, ActionEvent.CTRL_MASK));
		jMenuFileQuit.setMnemonic( 'Q' );
		setLocation(new java.awt.Point(0, 0));
		if( mauve.isApplication() )
			setJMenuBar(jMenuBar1);
//		getContentPane().setLayout(null);
		setTitle("Mauve - Genome Alignment Visualization");

		jMenuBar1.add(jMenuFile);
		jMenuFile.add(jMenuFileOpen);
		jMenuFile.add(jMenuFileAlign);
		jMenuFile.add(jMenuFilePrint);
		jMenuFile.add(jMenuFileClose);
		jMenuFile.add(jMenuFileSeparator1);
		jMenuFile.add(jMenuFileQuit);
		
		setSize(new java.awt.Dimension(791, 365));

		// event handling
		addWindowListener(new java.awt.event.WindowAdapter() {
			public void windowClosing(java.awt.event.WindowEvent e) {
				thisWindowClosing(e);
			}
		});

// END GENERATED CODE
		jMenuFileOpen.addActionListener(this);
		jMenuFileAlign.addActionListener(this);
		jMenuFileClose.addActionListener(this);
		jMenuFilePrint.addActionListener(this);
		jMenuFileQuit.addActionListener(this);
		
		status_bar = new LCBStatusBar();
		getContentPane().add( status_bar, BorderLayout.SOUTH );

		// set up key bindings
		jMenuBar1.getInputMap( JComponent.WHEN_IN_FOCUSED_WINDOW ).put( KeyStroke.getKeyStroke( "ctrl P" ), "Print" );
		jMenuBar1.getInputMap( JComponent.WHEN_IN_FOCUSED_WINDOW ).put( KeyStroke.getKeyStroke( "ctrl O" ), "Open" );
		jMenuBar1.getInputMap( JComponent.WHEN_IN_FOCUSED_WINDOW ).put( KeyStroke.getKeyStroke( "ctrl W" ), "Close" );
		jMenuBar1.getInputMap( JComponent.WHEN_IN_FOCUSED_WINDOW ).put( KeyStroke.getKeyStroke( "ctrl Q" ), "Quit" );
		jMenuBar1.getActionMap().put("Print", new GenericAction( this, "Print" ) );
		jMenuBar1.getActionMap().put("Open", new GenericAction( this, "Open" ) );
		jMenuBar1.getActionMap().put("Close", new GenericAction( this, "Close" ) );
		jMenuBar1.getActionMap().put("Quit", new GenericAction( this, "Quit" ) );

//		getContentPane().setLayout( new GridLayout( 1, 1 ) );
//		getContentPane().add( rrpanel, BorderLayout.NORTH );
		
/*		
		Image icon;
		try {
		    URL iconURL = new URL("file://icon.gif");
		    icon = getImage(iconURL);
		} catch (MalformedURLException murle) {
		}
		
		setIconImage(icon);
*/
        new FileDrop( System.out, getContentPane(), /*dragBorder,*/ new FileDrop.Listener()
        {   public void filesDropped( java.io.File[] files )
            {   for( int i = 0; i < files.length; i++ )
                {   
					loadFile( files[i] );
                }   // end for: through each dropped file
            }   // end filesDropped
        }); // end FileDrop.Listener

	}

	class GenericAction extends AbstractAction {

		GenericAction( ActionListener al, String command ){
			super( command );
			this.al = al;
			putValue( ACTION_COMMAND_KEY, command );
		}

		ActionListener al;
		public void actionPerformed(ActionEvent e){
			al.actionPerformed( e );
		}
	}
  
  	private boolean mShown = false;
  	
  	public void actionPerformed(ActionEvent ae) {
  		JMenuItem source = null;	// try to cast the source to a JMenuItem
  		try{
			source = (JMenuItem)(ae.getSource());
  		}catch( Exception e ){}
		if( source == jMenuFileQuit || ae.getActionCommand().equals( "Quit" ) ){
			setVisible(false);
			dispose();
			System.exit(0);
		}
		if( source == jMenuFileOpen || ae.getActionCommand().equals( "Open" ) ){
			doFileOpen();
		}
		if( source == jMenuFileClose || ae.getActionCommand().equals( "Close" ) ){
			thisWindowClosing( null );
		}
		if( source == jMenuFileAlign ){
			doAlign();
		}
		if( source == jMenuFilePrint || ae.getActionCommand().equals( "Print" ) ){
			PrintUtilities.printComponent( rrpanel );
		}
  	}
  	
	public void addNotify() {
		super.addNotify();
		
		if (mShown)
			return;
			
		// resize frame to account for menubar
		JMenuBar jMenuBar = getJMenuBar();
		if (jMenuBar != null) {
			int jMenuBarHeight = jMenuBar.getPreferredSize().height;
			Dimension dimension = getSize();
			dimension.height += jMenuBarHeight;
			setSize(dimension);
		}

		mShown = true;
	}

	// Close the window when the close box is clicked
	void thisWindowClosing(java.awt.event.WindowEvent e) {
		if( mauve.getFrameCount() > 1 || rrpanel == null ){
			setVisible(false);
			dispose();
			mauve.removeFrame( this );
		}else{
			getContentPane().remove( rrpanel );
			removeKeyListener( rrpanel );
			rrpanel = null;
			jMenuBar1.remove( toolbar );
			toolbar = null;
			jMenuBar1.validate();
			jMenuBar1.repaint();
			getContentPane().repaint();
			setTitle( "Mauve - Genome Alignment Visualization" );
			jMenuFilePrint.setEnabled(false);
		}
	}
	
	void loadFile( File rr_file ) {
		// if this frame's rrpanel is null, load the data.
		// otherwise create a new frame for the data.
		if( rrpanel == null )
		{
			toolbar = new JToolBar();
			jMenuBar1.add( toolbar );
			rrpanel = new DnDRearrangementPanel( this );
			getContentPane().add( rrpanel );
			addKeyListener( rrpanel );
			//this is where a real application would open the file.
			System.out.println( "Opening: " + rr_file.getPath() + ".\n" );
			
			RandomAccessFile xmfa_file;
			try{
				xmfa_file = new RandomAccessFile( rr_file, "r" );
				if( xmfa_file == null )
					return;
			}catch( Exception e ){
				System.err.println( "Could not open: " + rr_file.getPath() );
				return;
			}
			setTitle( "Mauve - Reading " + rr_file.getName() );
			try{
				rrpanel.ReadRearrangementData( xmfa_file );
				validate();
				toFront();
				/* go into visualization mode: set the title and enable the print option */
				setTitle( "Mauve - " + rr_file.getName() );
				jMenuFilePrint.setEnabled(true);
			}catch( Exception e ){ 
			e.printStackTrace(); 
			}

/*	        new FileDrop( System.out, rrpanel, new FileDrop.Listener()
	        {   public void filesDropped( java.io.File[] files )
	            {   for( int i = 0; i < files.length; i++ )
	                {   
						loadFile( files[i] );
	                }   // end for: through each dropped file
	            }   // end filesDropped
	        }); // end FileDrop.Listener
*/
		}else{
			MauveFrame frame = new MauveFrame( mauve );
			frame.fc = new JFileChooser( fc.getCurrentDirectory(), fc.getFileSystemView() );
			try{
				frame.initComponents();
				frame.setVisible(true);
				frame.loadFile( rr_file );
				mauve.addFrame( frame );
			}catch( Exception e  ){
				e.printStackTrace();
			}
		}
	}


	void loadURL( String rr_data_location ) {
		// if this frame's rrpanel is null, load the data.
		// otherwise create a new frame for the data.
		if( rrpanel == null )
		{
			try {
			    URL url = new URL( rr_data_location );
			    URLConnection connection = url.openConnection();
			    
			    InputStream is = connection.getInputStream();
			    InputStreamReader isr = new InputStreamReader(is);
			    BufferedReader match_reader = new BufferedReader(isr);
			    
			    toolbar = new JToolBar();
				jMenuBar1.add( toolbar );
				rrpanel = new DnDRearrangementPanel( this );
				getContentPane().add( rrpanel );
				addKeyListener( rrpanel );
				//this is where a real application would open the file.
				System.out.println( "Opening: " + rr_data_location + "\n" );
				
				setTitle( "Mauve - Loading " + rr_data_location );
				try{
//					rrpanel.ReadRearrangementData( match_reader );
					System.err.println("Reading data from URLs is broken\n");
					validate();
					toFront();
					match_reader.close();
					/* go into visualization mode: set the title and enable the print option */
					setTitle( "Mauve - " + rr_data_location );
					jMenuFilePrint.setEnabled(true);
				}catch( Exception e ){ e.printStackTrace(); }
			    match_reader.close();
			    
			} catch (MalformedURLException mue) {
				mue.printStackTrace();
			} catch (IOException ioe) {
				ioe.printStackTrace();
			}

		}else{
			MauveFrame frame = new MauveFrame( mauve );
			try{
				frame.initComponents();
				frame.setVisible(true);
				frame.loadURL( rr_data_location );
				mauve.addFrame( frame );
			}catch( Exception e  ){
				e.printStackTrace();
			}
		}
	}


	void doAlign() {
		try{
			if( alignFrame == null ){
				alignFrame = new AlignFrame();
				alignFrame.initComponents();
				alignFrame.alignButton.addActionListener(new java.awt.event.ActionListener() {
					public void actionPerformed(java.awt.event.ActionEvent e) {
						alignButtonActionPerformed(e);
					}
				});
			}
			alignFrame.setVisible( true );
//				alignFrame.setVisible(false);
//				alignFrame.dispose();
		}catch( Exception e )
		{}
	}

	public void alignButtonActionPerformed(java.awt.event.ActionEvent e) {
		Vector cmd_vec = new Vector();
		String read_filename = null;
		String cur_cmd;
		boolean detect_lcbs = true;
		String os_type = System.getProperty( "os.name" );
		
		System.out.println( "OS name is: " + os_type );
		if( os_type.startsWith( "Windows" ) )
			cmd_vec.addElement( "mauveAligner" );
		else
			cmd_vec.addElement( "./mauveAligner" );

		if( alignFrame.getSeedSize() > 0 ){
			cmd_vec.addElement( "-m" );
			cur_cmd = Integer.toString(alignFrame.getSeedSize());
			cmd_vec.addElement( cur_cmd );
		}

		// get a good output file name
		String output_file = alignFrame.getOutput();
		cmd_vec.addElement( "-o" );
		try{
			if( output_file.length() > 0 ){
				File ofile = new File( output_file );
				if( ofile.isDirectory() ){
					ofile = File.createTempFile( "mauve", ".mln", ofile );
					output_file = ofile.getPath();
				}
			}else{
				File ofile = File.createTempFile( "mauve", ".mln" );
				output_file = ofile.getPath();
			}
		}catch( IOException ioe ){
			ioe.printStackTrace();
		}
		cmd_vec.addElement( output_file );

		read_filename = output_file;
		
		detect_lcbs = alignFrame.isLCBSearchEnabled();
		if( detect_lcbs ){

			if( !alignFrame.getRecursive() ){
				cmd_vec.addElement( "--no-recursion" );
			}

			if( alignFrame.getMinRange() != -1 ){
				cur_cmd = "--weight=";
				cur_cmd += Integer.toString(alignFrame.getMinRange());
				cmd_vec.addElement( cur_cmd );
			}
			
			if( alignFrame.getRecursive() ){
				if( alignFrame.getIslandSize() != -1 ){
					cur_cmd = "--island-size=";
					cur_cmd += Integer.toString(alignFrame.getIslandSize());
					cmd_vec.addElement( cur_cmd );
					
					cur_cmd = "--island-output=" + output_file + ".islands";
					cmd_vec.addElement( cur_cmd );
				}
				
				if( alignFrame.getBackboneSize() != -1 &&
					alignFrame.getBackboneGapSize() != -1 ){
					cur_cmd = "--backbone-size=";
					cur_cmd += Integer.toString(alignFrame.getBackboneSize());
					cmd_vec.addElement( cur_cmd );

					cur_cmd = "--max-backbone-gap=";
					cur_cmd += Integer.toString(alignFrame.getBackboneGapSize());
					cmd_vec.addElement( cur_cmd );
					
					cur_cmd = "--backbone-output=" + output_file + ".backbone";
					cmd_vec.addElement( cur_cmd );
				}
				// make an identity matrix file name
				cur_cmd = "--id-matrix="+ output_file + ".id_matrix";
				cmd_vec.addElement( cur_cmd );

				// make an alignment file name
				cur_cmd = "--output-alignment=" + output_file + ".alignment";
				cmd_vec.addElement( cur_cmd );
				
				read_filename = output_file + ".alignment";
			}

			// make a guide tree file name
			cur_cmd = "--output-guide-tree=" + output_file + ".guide_tree";
			cmd_vec.addElement( cur_cmd );
		}else if( !detect_lcbs ){
			cur_cmd = "--mums";
			cmd_vec.addElement( cur_cmd );
			cur_cmd = "--eliminate-inclusions";
			cmd_vec.addElement( cur_cmd );
		}

		String[] sequences = alignFrame.getSequences();
		for( int seqI = 0; seqI < sequences.length; seqI++ ){
			cmd_vec.addElement( sequences[ seqI ] );
			if( sequences.length > 1 ){
				cur_cmd = sequences[ seqI ] + ".sml";
				cmd_vec.addElement( cur_cmd );
			}
		}

		String[] mauve_cmd = new String[ cmd_vec.size() ];
		mauve_cmd = (String[])(cmd_vec.toArray( mauve_cmd ));
		String mauve_str = new String();
		for( int cmdI = 0; cmdI < mauve_cmd.length; cmdI++ )
			mauve_str += mauve_cmd[ cmdI ] + " ";

		int retcode = -1;
		try{
			System.out.println( "Executing: " );
			System.out.println( mauve_str );
			Process align_proc = Runtime.getRuntime().exec( mauve_cmd );
			InputStreamReader err = new InputStreamReader(align_proc.getErrorStream());
			InputStreamReader in = new InputStreamReader(align_proc.getInputStream());
			while( true ){
				try{
					while( err.ready() )
						System.err.write( err.read() );
					while( in.ready() )
						System.out.write( in.read() );
					Thread.sleep( 100 );
					retcode = align_proc.exitValue();
					break;
				}catch(IllegalThreadStateException itse ){
				
				}
			}

//			int retcode = align_proc.waitFor();
			System.out.println( "Exited with code: " + retcode );
		}catch( Exception ioe ){
			ioe.printStackTrace();
		}

		if( retcode == 0 ){
			alignFrame.setVisible(false);
//			alignFrame.dispose();
//			alignFrame = null;
			if( rrpanel == null ){
				loadFile( new File( read_filename ) );
			}else{
				MauveFrame frame = new MauveFrame( mauve );
//				frame.alignFrame = alignFrame.clone();
				frame.fc = new JFileChooser( fc.getCurrentDirectory(), fc.getFileSystemView() );
				try{
					frame.initComponents();
					frame.setVisible(true);
					frame.loadFile( new File( read_filename ) );
					mauve.addFrame( frame );
				}catch( Exception e2  ){
					e2.printStackTrace();
				}
			}
		}else{
			JOptionPane.showMessageDialog(null, "mauveAligner exited with an error code.  Check the log window for details", "An error occurred", JOptionPane.ERROR_MESSAGE);
		}
	}

	void doFileOpen() {
		int returnVal = fc.showOpenDialog(this);
		
		if (returnVal == JFileChooser.APPROVE_OPTION) {
			File rr_file = fc.getSelectedFile();
			loadFile( rr_file );
		}
	}
	
	public void propertyChange( PropertyChangeEvent evt ){
		System.out.println( "proppity: " + evt.getPropertyName() );
	}
	
	/**
	 * This class displays statistics about LCBs in a status bar
	 */
	class LCBStatusBar extends JPanel {
		JLabel hint_bar = new JLabel( " " );
		JLabel LCB_length_status = new JLabel( " " );
		JLabel LCB_weight_status = new JLabel( " " );

		LCBStatusBar(){

			setLayout( new BoxLayout( this, BoxLayout.X_AXIS ) );
			setMaximumSize( new Dimension( Short.MAX_VALUE, 15 ) );
			setBorder( BorderFactory.createEtchedBorder() );

			add( hint_bar );
			add( Box.createHorizontalGlue() );

			LCB_length_status.setBorder( BorderFactory.createLoweredBevelBorder() );
			Dimension pref_dim = LCB_length_status.getPreferredSize();
			pref_dim.setSize( 150, pref_dim.getHeight() );
			LCB_length_status.setPreferredSize( pref_dim );
			add( LCB_length_status );

			add( Box.createHorizontalStrut(2) );

			LCB_weight_status.setBorder( BorderFactory.createLoweredBevelBorder() );
			pref_dim = LCB_weight_status.getPreferredSize();
			pref_dim.setSize( 150, pref_dim.getHeight() );
			LCB_weight_status.setPreferredSize( pref_dim );
			add( LCB_weight_status );

			add( Box.createHorizontalStrut(1) );
		}
		
		public void setWeightStatus( long weight ){
			LCB_weight_status.setText( " LCB weight: " + weight );
		}
		public void clearWeightStatus(){
			LCB_weight_status.setText( " " );
		}
		
		public void setLengthStatus( long length ){
			LCB_length_status.setText( " LCB length: " + length );
		}
		public void clearLengthStatus(){
			LCB_length_status.setText( " " );
		}
		
		public void setHint( String hint ){
			hint_bar.setText( hint );
		}

		public void clearHint(){
			hint_bar.setText( "" );
		}

	}

}
