package org.gel.mauve;

import java.awt.*;
import javax.swing.*;
import java.awt.print.*;

/** A simple utility class that lets you very simply print
 *  an arbitrary component. Just pass the component to the
 *  PrintUtilities.printComponent. The component you want to
 *  print doesn't need a print method and doesn't have to
 *  implement any interface or do anything special at all.
 *  <P>
 *  If you are going to be printing many times, it is marginally more 
 *  efficient to first do the following:
 *  <PRE>
 *    PrintUtilities printHelper = new PrintUtilities(theComponent);
 *  </PRE>
 *  then later do printHelper.print(). But this is a very tiny
 *  difference, so in most cases just do the simpler
 *  PrintUtilities.printComponent(componentToBePrinted).
 *
 *  7/99 Marty Hall, http://www.apl.jhu.edu/~hall/java/
 *  May be freely used or adapted.
 */

public class PrintUtilities implements Printable {
  private Component componentToBePrinted;

  public static void printComponent(Component c) {
    new PrintUtilities(c).print();
  }
  
  public PrintUtilities(Component componentToBePrinted) {
    this.componentToBePrinted = componentToBePrinted;
  }
  
  public void print() {
    PrinterJob printJob = PrinterJob.getPrinterJob();
	PageFormat pformat = new PageFormat();
//	pformat.setOrientation( PageFormat.LANDSCAPE );
    printJob.setPrintable(this, pformat);
    if (printJob.printDialog())
      try {
        printJob.print();
      } catch(PrinterException pe) {
        System.out.println("Error printing: " + pe);
      }
  }

  public int print(Graphics g, PageFormat pageFormat, int pageIndex) {
    if (pageIndex > 0) {
      return(NO_SUCH_PAGE);
    } else {
      Graphics2D g2d = (Graphics2D)g;
      System.out.println( "imageable x: " + pageFormat.getImageableX() + "imageable y: " + pageFormat.getImageableY() );
      g2d.translate(pageFormat.getImageableX(), pageFormat.getImageableY());
      disableDoubleBuffering(componentToBePrinted);
      double dpi = 600;
      double scaling = ( pageFormat.getImageableWidth() / 72 ) * dpi;
      Dimension original_dim = componentToBePrinted.getSize();
      Dimension dim = componentToBePrinted.getSize();
      dim.height = dim.height * (int)( scaling / dim.width );
      dim.width = (int)scaling;
      System.out.println( "New dim: " + dim.toString() );
      componentToBePrinted.setSize( dim );
//      g2d.transform( new java.awt.geom.AffineTransform( pageFormat.getMatrix() ) );
      java.awt.geom.AffineTransform at = new java.awt.geom.AffineTransform( 72d / dpi, 0, 0, 72d / dpi, 0, 0  );
      g2d.transform( at );
      try{
	      RearrangementPanel rrp = (RearrangementPanel)componentToBePrinted;
	      rrp.layoutDisplay();
	  }catch(Exception e ){}
      componentToBePrinted.paint( g2d );
      componentToBePrinted.setSize( original_dim );
      enableDoubleBuffering(componentToBePrinted);
      return(PAGE_EXISTS);
    }
  }

  /** The speed and quality of printing suffers dramatically if
   *  any of the containers have double buffering turned on.
   *  So this turns if off globally.
   *  @see enableDoubleBuffering
   */
  public static void disableDoubleBuffering(Component c) {
    RepaintManager currentManager = RepaintManager.currentManager(c);
    currentManager.setDoubleBufferingEnabled(false);
  }

  /** Re-enables double buffering globally. */
  
  public static void enableDoubleBuffering(Component c) {
    RepaintManager currentManager = RepaintManager.currentManager(c);
    currentManager.setDoubleBufferingEnabled(true);
  }
}
