package org.gel.mauve;


import java.awt.event.*;
import java.applet.Applet;
import java.net.*;
import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.image.*;
import javax.swing.*;

public class RRSequencePanel extends JPanel implements MouseListener, MouseMotionListener {
	Vector boxes;				//*< A list of all the rearrangement boxes contained by this panel */
	Match[] m_matches;		//*< A list of the matches sorted on the sequence this panel is displaying */
	int seq;				//*< The index of the sequence this panel displays */
	long start_view;		//*< The first coordinate currently visible */
	long view_length;		//*< The length of visible coordinates */
	MatchStartCompare m_compare;	//*< The comparator being used to sort the match Vector */
	String sequence_name;	//*< The name of the sequence in this panel */

	double box_fill_percent = 65;	//*< The percentage height of each rearrangement bar. The rest is used to draw highlighting.*/
	Color name_color = new Color( 128, 128, 128 );
	
	// used for drawing the rectangles.
	int first_match;
    Dimension match_off_dim;
    BufferedImage match_off_image;
    Graphics2D match_off_graphics;
//    Image match_off_image;
//    Graphics match_off_graphics;

    Dimension highlight_off_dim;
    Image highlight_off_image;
    Graphics highlight_off_graphics;
	static final Color highlight_color = new Color( 24, 24, 24 );

	long off_start_view;
	long off_view_len;	
	
	static final int MAX_POPUP_MATCHES = 15;	/**< The largest number of matches which will be shown in a popup menu */
	
	long sequence_length;	/**< The length of the sequence drawn by this panel */
	
	boolean draw_lcb_bounds = true;
	boolean fill_lcb_boxes = false;
	boolean draw_match_boxes = true;
	RearrangementPanel rrpanel;	/**< The rearrangement panel that contains this sequence panel */
	SimilarityIndex sim;
	
	public RRSequencePanel( RearrangementPanel rrpanel, Vector matches, int sequence, String sequenceName, long sequence_length, SimilarityIndex sim ) {
		seq = sequence;
		sequence_name = sequenceName;
		this.sequence_length = sequence_length;
		this.rrpanel = rrpanel;
		this.sim = sim;

		// initialize and sort the start position array
		m_matches = new Match[ matches.size() ];
		m_matches = (Match[]) matches.toArray( m_matches );
		m_compare = new MatchStartCompare( seq );
		m_compare.seq = seq;
		Arrays.sort( m_matches, m_compare );

		addMouseListener( this );
		addMouseMotionListener( this );
	}
	public void setSequence( int sequence ){
		seq = sequence;
		m_compare = new MatchStartCompare( seq );
		m_compare.seq = seq;
		Arrays.sort( m_matches, m_compare );
		match_off_graphics = null;
		highlight_off_graphics = null;
		formatBoxes();	// the image needs to change correspondingly
	}
	
	public void mouseDragged(MouseEvent e) {}
	public void mouseMoved(MouseEvent e){

		// set the LCB stats for this region
		int[] lcbs = getLCBPixelRange( e.getX(), e.getX() );
		int lcbI = 0;
		for( ; lcbI < lcbs.length; lcbI++ ){
			LCB cur_lcb = rrpanel.lcb_list[ lcbs[ 0 ] ];
			int y = e.getY() - box_top;
			if( (cur_lcb.reverse[ seq ] && y > half_height && y <= box_height) ||
				(!cur_lcb.reverse[ seq ] && y > 0 && y <= half_height) ){

				rrpanel.mauve_frame.status_bar.setWeightStatus( cur_lcb.weight );
				long length = cur_lcb.right_end[ seq ] - cur_lcb.left_end[ seq ];
				rrpanel.mauve_frame.status_bar.setLengthStatus( length );
				break;
			}
		}
		// check whether anything could be set, if not, clear the status
		if( lcbI == lcbs.length ){
			rrpanel.mauve_frame.status_bar.clearWeightStatus();
			rrpanel.mauve_frame.status_bar.clearLengthStatus();
		}
		
		// highlight the appropriate region in each sequence
		if( e.getY() >= box_top && e.getY() <= box_top + box_height ){
			// map the mouse position to sequence coordinates
			long[] seq_range = new long[2];
			pixelRangeToSequenceCoordinates( e.getX(), e.getX(), seq_range );
			if( seq_range[ 0 ] >= 0 && seq_range[ 0 ] < sequence_length )
				rrpanel.highlightRegions( seq, seq_range[ 0 ] );
		}
	}
 	
 	Vector highlight_coords = new Vector();
 	public void highlight( long seq_coord ){
 		highlight_coords.addElement( new Long( seq_coord ) );
 	}
 	public void clearHighlights(){
 		highlight_coords = new Vector();
	}

	// extra methods which aren't implemented
	public void mouseEntered(MouseEvent e) {}
	public void mouseExited(MouseEvent e) {}
	public void mousePressed(MouseEvent e) {}
	public void mouseReleased(MouseEvent e) {}

	public void mouseClicked(MouseEvent e) {
		if( e.isPopupTrigger() ) {
			System.out.println("Popup triggered\n");
			MatchPopupMenu new_popup = getPopup( e );
			new_popup.show( this, e.getX(), e.getY() );
		}		// do something here eventually!
		// popup ideas:  show popup menu with an "align sequences to" section which enumerates
		// each match in this box.
		// maybe also have a section for "show information on" with a list of each match.
		// i seem to be good at cutting work out for myself
		if( e.getClickCount() == 1 ){
			// if it is a control click then show the popup menu
			if( e.isControlDown() || ((e.getModifiers() & InputEvent.BUTTON3_MASK) != 0)){
				MatchPopupMenu new_popup = getPopup( e );
				add( new_popup );
				new_popup.show( this, e.getX(), e.getY() );
				
			}else{

//				RearrangementPanel rrpanel = (RearrangementPanel) (getParent());

				// first clear any previous highlights if shift isn't down
				if( ( e.getModifiers() & InputEvent.SHIFT_MASK ) == 0 ) {
					rrpanel.clearHighlights();
					if( rrpanel.lcb_list.length > 0 ){
						// clear the LCB status info
						rrpanel.mauve_frame.status_bar.clearWeightStatus();
						rrpanel.mauve_frame.status_bar.clearLengthStatus();
					}
				}
				
				if( sim != null ){
					// map the clicked site to a sequence coordinates
					long[] seq_range = new long[2];
					pixelRangeToSequenceCoordinates( e.getX(), e.getX(), seq_range );
					rrpanel.alignView( seq, seq_range[ 0 ] );
				}else{
					int[] match_range = new int[2];
					getMatchPixelRange( e.getX(), e.getX(), match_range );
					// highlight this and the other components of the match
					for( int matchI = match_range[0]; matchI < match_range[1]; matchI++ ){
						rrpanel.addHighlighted( m_matches[ matchI ] );
						m_matches[ matchI ].highlighted = true;
						if( m_matches[ matchI ].lcb >= 0 && rrpanel.lcb_list.length > 0 ){
							// set the LCB status info
							LCB cur_lcb = rrpanel.lcb_list[ m_matches[ matchI ].lcb ];
							rrpanel.mauve_frame.status_bar.setWeightStatus( cur_lcb.weight );
							long length = cur_lcb.right_end[ seq ] - cur_lcb.left_end[ seq ];
							rrpanel.mauve_frame.status_bar.setLengthStatus( length );
						}
					}
					rrpanel.repaintSeqPanels();
				}
			}
		}else if( e.getClickCount() == 2 ){
			// do something like move the annotation browser window onto this section of sequence.
			System.out.println( "doubleclick.net!" );
		}
	}

	MatchPopupMenu getPopup( MouseEvent e ){
		MatchPopupMenu pop_menu = new MatchPopupMenu((RearrangementPanel) getParent(), seq);

		int[] match_range = new int[2];
		getMatchPixelRange( e.getX(), e.getX(), match_range );
		if( match_range[1] - match_range[0] < MAX_POPUP_MATCHES ){
			for(int matchI = match_range[0]; matchI < match_range[1]; matchI++ )
				pop_menu.addMatch( m_matches[ matchI ] );
		}else{
			double pop_interval = (double)(match_range[1] - match_range[0]) / (double)MAX_POPUP_MATCHES;
			double popI = match_range[0];
			for(int matchI = match_range[0]; matchI < match_range[1]; matchI++ ){
				if( (int)popI == matchI )
					pop_menu.addMatch( m_matches[ matchI ] );
				popI += pop_interval;
			}
		}
		return pop_menu;
	}

	/**
	 * Converts a range of pixel coordinates to sequence coordinates
	 */
	protected void pixelRangeToSequenceCoordinates( int start_pixel, int end_pixel, long[] seq_range ){
		Dimension dim = getSize();
		double pixel_width = (double)view_length / (double)dim.width;
		seq_range[0] = (long) (start_pixel * pixel_width) + start_view;
		seq_range[1] = (long) ((end_pixel + 1) * pixel_width) + start_view - 1;
		seq_range[1] = seq_range[0] > seq_range[1] ? seq_range[0] : seq_range[1];
	}
	
	/**
	 * Finds all LCBs intersecting a specified range of pixels in the current view.
	 * @param start_pixel The first coordinate of the intersection range
	 * @param end_pixel The last coordinate of the intersection range
	 */
	protected int[] getLCBPixelRange( int start_pixel, int end_pixel ){
		Dimension dim = getSize();
		long[] seq_range = new long[ 2 ];
		pixelRangeToSequenceCoordinates( start_pixel, end_pixel, seq_range );
		return getLCBRange( seq_range[0], seq_range[1] );
	}

	/**
	 * Finds all LCBs that intersect the specified coordinate range in the current view
	 * @param start_coord 	The first coordinate of the intersection range
	 * @param end_coord 	The last coordinate of the intersection range
	 * @param lcbs  	An int array of the intersecting LCB indices
	 */
	protected int[] getLCBRange( long start_coord, long end_coord ){
		// count first, then add.  stupid java with no dynamic array support for builtin types
		int lcb_count = 0;
		for( int lcbI = 0; lcbI < rrpanel.lcb_list.length; lcbI++ ){
			long lend = rrpanel.lcb_list[ lcbI ].left_end[ seq ];
			long rend = rrpanel.lcb_list[ lcbI ].right_end[ seq ];
			// check for intersection
			if( rend >= start_coord && lend <= end_coord )
				lcb_count++;
		}
		int[] lcbs = new int[ lcb_count ];
		lcb_count = 0;
		for( int lcbI = 0; lcbI < rrpanel.lcb_list.length; lcbI++ ){
			long lend = rrpanel.lcb_list[ lcbI ].left_end[ seq ];
			long rend = rrpanel.lcb_list[ lcbI ].right_end[ seq ];
			// check for intersection
			if( rend >= start_coord && lend <= end_coord )
				lcbs[ lcb_count++ ] = rrpanel.lcb_list[ lcbI ].lcb_id;
		}
		return lcbs;
	}
	
	/**
	 * Finds all match indices which intersect with the specified range of pixels in the current
	 * view.
	 * @param start_pixel The first coordinate of the intersection range
	 * @param end_pixel The last coordinate of the intersection range
	 * @param match_range An int array with 2 elements.  The resulting range of intersecting
	 * match indices will be returned as the first and second elements in the array.
	 */
	protected void getMatchPixelRange( int start_pixel, int end_pixel, int[] match_range ){
		Dimension dim = getSize();
		long[] seq_range = new long[ 2 ];
		pixelRangeToSequenceCoordinates( start_pixel, end_pixel, seq_range );
		getMatchRange( seq_range[0], seq_range[1], match_range );
	}

	/**
	 * Provides a range of indexes into the m_matches array which contain matches intersecting
	 * with the specified range of coordinates
	 * @param start_coord 	The first coordinate of the intersection range
	 * @param end_coord 	The last coordinate of the intersection range
	 * @param match_range 	An int array with 2 elements.  The resulting range of intersecting
	 * match indices will be returned as the first and second elements in the array.
	 */
	protected void getMatchRange( long start_coord, long end_coord, int[] match_range ){
		if( m_matches.length == 0 ){
			match_range[ 0 ] = -1;
			match_range[ 1 ] = -1;
			return;
		}

		Match f_match = new Match( m_matches[0] );

		f_match.starts[ seq ] = start_coord;
		if( f_match.starts[ seq ] < 1 )
			f_match.starts[ seq ] = 1;

		int match_startI = Arrays.binarySearch( m_matches, f_match, m_compare );
		match_startI = match_startI >= 0 ? match_startI : -match_startI - 1;

		f_match.starts[ seq ] = end_coord + 1;
		int match_endI = Arrays.binarySearch( m_matches, f_match, m_compare );
		match_endI = match_endI >= 0 ? match_endI : -match_endI - 1;
		
		// now look backwards 1 match for an intersection
		if( match_startI > 0 ){
			Match prev_match = m_matches[ match_startI - 1 ];

			if( prev_match.starts[ seq ] + prev_match.lengths[ seq ] > start_coord &&
				prev_match.starts[ seq ] != Match.NO_MATCH )
				match_startI--;
		}
		match_range[0] = match_startI;
		match_range[1] = match_endI;
	}
	
	/**
	 * Set the range of visible matches.
	 * @param start The sequence coordinate to start displaying matches
	 * @param len The length of the range of matches to display starting at @param start
	 */
	public void setViewableRange( long start, long len ) {
		start_view = start;
		view_length = len;
	}
	
	/**
	 * Returns the RRSequencePanel's currently displayed coordinate range in the supplied array.
	 * The parameter @param view_array should be an array of longs with size 2.
	 * The first element will contain the start of the view range and the second element will contain
	 * the length of the view range.
	 */
	public void getViewableRange( long[] view_array) {
		view_array[0] = start_view;
		view_array[1] = view_length;
	}
	
	public int sequenceCoordinateToPixel( long pos ) {
		Dimension dim = getSize();
		pos -= start_view;
		pos *= dim.width;
		if( view_length == 0 )
			return 0;
		return (int)(pos / view_length);
	}
	
	public void update( Graphics g ) {
		paint( g );
	}

	public void paintComponent( Graphics g ) {
		paintRange( g, getSize(), start_view, view_length, true );
	}

	public void paint( Graphics g ) {
		paintRange( g, getSize(), start_view, view_length, true );
	}
	
	protected int boxTop( Dimension dim ){
		return (int)(((100d - box_fill_percent) / 200d) * dim.height) + 1;
	}
	
	protected int boxHeight( Dimension dim ){
		return (int)((box_fill_percent / 100d) * dim.height) + 1;
	}

	public void paintRange( Graphics graphics, Dimension dim, long start_c, long coord_len, boolean draw_highlighting ) {
	try{
		Graphics2D g = (Graphics2D)graphics;
		Image matchImage = null;
		Image highlightImage = null;
		if( dim.equals( getSize() ) && start_c == start_view && coord_len == view_length ){
		}else{
			formatBoxes( dim );
		}
		if( highlight_off_image == null || match_off_image == null )
			formatBoxes( dim );
		
		matchImage = match_off_image;
		highlightImage = highlight_off_image;

		int box_top = boxTop( dim );
		int box_height = boxHeight( dim );
		if( !draw_highlighting ) {
			// just draw the matches so they cover the highlighting
			box_top = 0;
			box_height = dim.height;
		}

		g.setColor( RearrangementPanel.bg_color );
		g.fillRect( 0, 0, dim.width, dim.height );

		// draw the highlighting
	    g.drawImage(highlightImage, 0, 0, dim.width, 
	    	dim.height, 0, 0, highlightImage.getWidth(null), 
	    	highlightImage.getHeight(null), null);

		// draw the matches
	    g.drawImage(matchImage, 0, box_top, dim.width, 
	    	box_top + box_height, 0, 0, matchImage.getWidth(null), 
	    	matchImage.getHeight(null), null);

		// draw the center line
		g.setColor( Color.black );
		g.drawLine( 0, dim.height / 2, dim.width, dim.height / 2 );

		
		// fill any unused area with diagonal lines
/*		if( start_view < 1 ){
			int last_right = sequenceCoordinateToPixel( 1 ) - 1;
			paintDiagonalLineFill( g, 0, 0, last_right, dim.height );
		}

		if( start_view + view_length > sequence_length ){
			int last_pix = sequenceCoordinateToPixel( sequence_length );
			paintDiagonalLineFill( g, last_pix, 0, dim.width, dim.height );
		}
*/		
		// draw any highlights
		g.setColor( Color.black );
		for( int highI = 0; highI < highlight_coords.size(); highI++ ){
			long high_c = ((Long)highlight_coords.elementAt( highI )).longValue();
			if( high_c < start_view ){
				// draw an arrow?
				continue;
			}
			if( high_c > start_view + view_length ){
				// draw an arrow?
				continue;
			}
			int high_pixel = sequenceCoordinateToPixel( high_c );
			g.drawLine( high_pixel, 0, high_pixel, dim.height );
		}

		// draw the name of the sequence below the boxes.
		// determine what font size to draw the sequence label with
		int name_space = dim.height - box_top - box_height;
		g.setColor( name_color );
		if( name_space < 22 ){
			Font f = new Font( "Default", Font.PLAIN, 10 );
			g.setFont( f );
			g.drawString( sequence_name, 10, dim.height - 2 );
		}else{
			Font f = new Font( "Default", Font.PLAIN, (2 * name_space ) / 3 );
			g.setFont( f );
			g.drawString( sequence_name, 10, dim.height - (name_space / 4) );
		}

		}catch(Exception e){
			e.printStackTrace();
		}
	}
	
	protected void paintDiagonalLineFill( Graphics graphics, int left, int top, int right, int bottom ){
		Graphics2D g = (Graphics2D)graphics;
		Color lt_gray = Color.getHSBColor(0, 0, .83f);
		Color dk_gray = Color.getHSBColor(0, 0, .85f);
		int line_width = 20;
		Color cur_color = lt_gray;
		g.setColor( cur_color );
		// along the top edge first
		for( int pixelI = right - 1; pixelI > left - 1; pixelI-- ){
			int len = bottom - top < pixelI - left ? bottom - top : pixelI - left;
			g.drawLine( pixelI, top, pixelI - len, top + len );
			if( (right - pixelI) % line_width == 0 ){
				if( cur_color == lt_gray )
					cur_color = dk_gray;
				else
					cur_color = lt_gray;
				g.setColor( cur_color );
			}
		} 
		cur_color = lt_gray;
		g.setColor( cur_color );
		// now along the right side
		for( int pixelI = top; pixelI < bottom; pixelI++ ){
			int len = bottom - pixelI < right - left ? bottom - pixelI : right - left;
			g.drawLine( right, pixelI, right - len, pixelI + len );
			if( pixelI % line_width == 0 ){
				if( cur_color == lt_gray )
					cur_color = dk_gray;
				else
					cur_color = lt_gray;
				g.setColor( cur_color );
			}
		} 

	}
	
	public boolean getDrawLCBbounds(){
		return draw_lcb_bounds;
	}
	
	public void setDrawLCBbounds( boolean draw_lcb_bounds ){
		this.draw_lcb_bounds = draw_lcb_bounds;
	}

	public boolean getFillLCBboxes(){
		return fill_lcb_boxes;
	}

	public void setFillLCBboxes( boolean fill_lcb_boxes ){
		this.fill_lcb_boxes = fill_lcb_boxes;
	}
	
	public boolean getDrawMatches(){
		return draw_match_boxes;
	}

	public void setDrawMatches( boolean draw_match_boxes ){
		this.draw_match_boxes = draw_match_boxes;
	}
	
	protected void paintRange( Image matchImage, Image highlightImage, long start_c, long coord_len ){
	
	}
	
	double pixel_width;
	int box_height;
	int half_height;
	int box_top;
	/**
	 * Organize the matches into visual boxes to be displayed on screen.  formatBoxes
	 * will coalesce small matches into a single box to be displayed.
	 */
	protected void formatBoxes(){
		Dimension dim = getSize();
		formatBoxes( dim );
	}
	protected void formatBoxes( Dimension dim ){
try{
		if( dim.height <= 10 || dim.width <= 10 )
			return;

		// locate the first and last matches to draw in the array.
		long end_view = start_view + view_length - 1;
		
		pixel_width = (double)view_length / (double)dim.width;
		box_height = (int)(( box_fill_percent * dim.height ) / 100d );
		half_height = box_height / 2;
		box_top = (dim.height - box_height) / 2;
		
		// Update the offscreen graphics context if necessary
		// It should be 3 times the size of the displayable screen to facilitate
		// quick scrolls
		if ((match_off_graphics == null)
		 || (dim.width != match_off_dim.width)
		 || (dim.height != match_off_dim.height)) {
		    match_off_dim = new Dimension( dim.width, box_height );
		    match_off_image = new BufferedImage( match_off_dim.width, match_off_dim.height, BufferedImage.TYPE_INT_ARGB );
		    match_off_graphics = (Graphics2D)match_off_image.getGraphics();
		    
		}else{
			// otherwise erase it
			match_off_graphics.setColor( RearrangementPanel.bg_color );
			match_off_graphics.fillRect( 0, 0, match_off_dim.width, match_off_dim.height );
		}

		if ((highlight_off_graphics == null)
		 || (dim.width != highlight_off_dim.width)
		 || (dim.height != match_off_dim.height)) {
		    highlight_off_dim = new Dimension( dim.width, 2 );
		    highlight_off_image = createImage( highlight_off_dim.width, highlight_off_dim.height );
		    highlight_off_graphics = highlight_off_image.getGraphics();
		    highlight_off_graphics.setColor( highlight_color );
		}else{
			// otherwise erase it
			highlight_off_graphics.setColor( RearrangementPanel.bg_color );
			highlight_off_graphics.fillRect( 0, 0, highlight_off_dim.width, highlight_off_dim.height );
		    highlight_off_graphics.setColor( highlight_color );
		}

		// set default pen width	
		match_off_graphics.setStroke( new BasicStroke( 0f ) );

		// LCB drawing parameters
		double lcb_boundary_width = 3d;
		double arc_size = (double)half_height / 6d;
		double half_pen_width = lcb_boundary_width / 2d;
		LCB[] lcb_list = rrpanel.lcb_list;
		int lcbI = 0;

		// draw white backgrounds for LCB rectangles
		if( draw_lcb_bounds ){
			// set the pen width
			match_off_graphics.setColor( Color.white );
			for( lcbI = 0; lcbI < lcb_list.length; lcbI++ ){
				// skip this LCB if it's out of viewing range
				if( start_view > lcb_list[ lcbI ].right_end[ seq ] )
					continue;
				if( end_view < lcb_list[ lcbI ].left_end[ seq ] )
					continue;

				int cur_start_pixel = (int)((double)( lcb_list[ lcbI ].left_end[ seq ] - start_view ) / pixel_width);
				double lcb_length = lcb_list[ lcbI ].right_end[ seq ] - lcb_list[ lcbI ].left_end[ seq ];
				int cur_pixel_width = (int) Math.ceil( lcb_length / pixel_width );
				int lcb_top = lcb_list[ lcbI ].reverse[ seq ] ? half_height : 0;
				
				java.awt.geom.RoundRectangle2D.Double lcb_rect = new java.awt.geom.RoundRectangle2D.Double( cur_start_pixel + half_pen_width, lcb_top + half_pen_width, cur_pixel_width - lcb_boundary_width, half_height - lcb_boundary_width, arc_size, arc_size );
				match_off_graphics.fill( lcb_rect );
			}
		}


		// draw the sequence similarity values if sim isn't null!
		try{
		long starttime = System.currentTimeMillis();
		if( draw_match_boxes && sim != null ){

			// find the left-most LCB in the displayed list
			lcbI = -1;
			for( int left_lcb = 0; left_lcb < lcb_list.length; left_lcb++ ){
				if( lcb_list[ left_lcb ].left_adjacency[ seq ] < 0 )
					lcbI = left_lcb;
			}
			// find the left-most in the deleted list
			int del_lcbI = -1;
			for( int left_lcb = 0; left_lcb < rrpanel.del_lcb_list.length; left_lcb++ ){
				if( rrpanel.del_lcb_list[ left_lcb ].left_adjacency[ seq ] < 0 )
					del_lcbI = left_lcb;
			}

			match_off_graphics.setColor( Color.gray );
			double sim_height = half_height - lcb_boundary_width;
			boolean reverse = lcb_list[ lcbI ].reverse[ seq ];

			for( int pixelI = 0; pixelI < dim.width; pixelI++ ){
				long seq_left = (long)((double)pixelI * pixel_width) + start_view;
				long seq_right = (long)((double)(pixelI+1) * pixel_width) + start_view;
				seq_left = seq_left < 0 ? 0 : seq_left;
				seq_right = seq_right < sequence_length ? seq_right : sequence_length;
				if( seq_left > sequence_length )
					continue;
				if( seq_right < 0 )
					continue;
				if( seq_right == sequence_length )
					System.out.print("");	// catch this and debuggeroni
				// if we've moved beyond the current lcb, move to the next one
				while( lcbI >= 0 && seq_left > lcb_list[ lcbI ].right_end[ seq ] )
					lcbI = lcb_list[ lcbI ].right_adjacency[ seq ];
				// scan up to the next deleted LCB
				while( del_lcbI >= 0 && seq_left > rrpanel.del_lcb_list[ del_lcbI ].right_end[ seq ] )
					del_lcbI = rrpanel.del_lcb_list[ del_lcbI ].right_adjacency[ seq ];
				// don't draw anything if we're not within a valid LCB
				if( lcbI < 0 || seq_right < lcb_list[ lcbI ].left_end[ seq ] )
					continue;
				if( del_lcbI >= 0 && seq_right >= rrpanel.del_lcb_list[ del_lcbI ].left_end[ seq ] )
					continue;
				reverse = lcb_list[ lcbI ].reverse[ seq ];
				double s = sim.getSimilarityByRange( seq_left, seq_right );
				// normalize to a box_height
				double height = (((double)s + 127d)/256d * sim_height);
				double match_top = reverse ? half_height + half_pen_width : half_pen_width + sim_height - height;
				java.awt.geom.Rectangle2D.Double match_rect;
				match_rect = new java.awt.geom.Rectangle2D.Double( pixelI, match_top, 1, height );
				match_off_graphics.fill( match_rect );
			}
			
		}else if( draw_match_boxes ){	// draw boxes for matches instead of similarity values
		
			int[] match_range = new int[2];
			getMatchRange( start_view, end_view, match_range );
			first_match = match_range[0];
			long last_match = match_range[1];

			for( int directionI = 0; directionI < 2; directionI++ ){
				boolean reverse_matches = false;
				if( directionI == 1 )
					reverse_matches = true;
				int prev_start_pixel = -1;
				int prev_pixel_width = 0;
				int cur_start_pixel = 0;
				int cur_pixel_width = 0;
				int end_pixel = 0;
				
				double match_top = reverse_matches ? half_height + half_pen_width : half_pen_width;

				Color prev_color = null;
				boolean prev_reverse = false;
				for( int matchI = first_match; matchI < last_match; matchI++ ){
					Match cur_match = m_matches[ matchI ];
					// don't draw the match if it's not on the current strand
					// or if it's part of a 'disabled' LCB.
					if( cur_match.reverse[ seq ] == reverse_matches && cur_match.lcb >= 0 ){
						cur_start_pixel = (int)((double)( cur_match.starts[ seq ] - start_view ) / pixel_width);
						cur_pixel_width = (int) Math.ceil( (double)cur_match.lengths[ seq ] / pixel_width );
						if( cur_match.highlighted )
							if( reverse_matches )
								highlight_off_graphics.fillRect( cur_start_pixel, 1, cur_pixel_width, 1 );
							else
								highlight_off_graphics.fillRect( cur_start_pixel, 0, cur_pixel_width, 1 );
							
						if( cur_start_pixel > prev_start_pixel ){
							end_pixel = prev_start_pixel + prev_pixel_width - 1;
							end_pixel = cur_start_pixel < end_pixel ? cur_start_pixel : end_pixel;
							match_off_graphics.setColor( prev_color );
							java.awt.geom.Rectangle2D.Double match_rect = new java.awt.geom.Rectangle2D.Double( prev_start_pixel, match_top, end_pixel - prev_start_pixel + 1, half_height - lcb_boundary_width );
							match_off_graphics.fill( match_rect );
						}
						
						prev_start_pixel = cur_start_pixel;
						prev_pixel_width = cur_pixel_width;
						prev_reverse = cur_match.reverse[ seq ];
						prev_color = m_matches[ matchI ].color;
					}
				}
				
				// draw the last match.
				match_off_graphics.setColor( prev_color );
				if( prev_reverse ){
					match_off_graphics.fillRect( prev_start_pixel, half_height, prev_pixel_width, box_height );
				}else{
					match_off_graphics.fillRect( prev_start_pixel, 0, prev_pixel_width, half_height );
				}
			}	// for directionI 

		}	// if( draw_match_boxes )
		}catch(Exception e ){
			e.printStackTrace();
		}
		
		// draw the LCB bounding rectangles
		if( draw_lcb_bounds ){
			// set the pen width
			match_off_graphics.setStroke( new BasicStroke( (float)lcb_boundary_width ) );
			for( lcbI = 0; lcbI < lcb_list.length; lcbI++ ){
				if( start_view > lcb_list[ lcbI ].right_end[ seq ] )
					continue;
				if( end_view < lcb_list[ lcbI ].left_end[ seq ] )
					continue;

				int cur_start_pixel = (int)((double)( lcb_list[ lcbI ].left_end[ seq ] - start_view ) / pixel_width);
				double lcb_length = lcb_list[ lcbI ].right_end[ seq ] - lcb_list[ lcbI ].left_end[ seq ];
				int cur_pixel_width = (int) Math.ceil( lcb_length / pixel_width );
				int lcb_top = lcb_list[ lcbI ].reverse[ seq ] ? half_height : 0;
				
				if( rrpanel.lcb_list[ lcbI ].color != null )
					match_off_graphics.setColor( rrpanel.lcb_list[ lcbI ].color );

				java.awt.geom.RoundRectangle2D.Double lcb_rect = new java.awt.geom.RoundRectangle2D.Double( cur_start_pixel + half_pen_width, lcb_top + half_pen_width, cur_pixel_width - lcb_boundary_width, half_height - lcb_boundary_width, arc_size, arc_size );

				match_off_graphics.draw( lcb_rect );
				if( fill_lcb_boxes )
					match_off_graphics.fill( lcb_rect );
			}

			// now draw deleted LCBs in obnoxious yellow-orange
			match_off_graphics.setColor( Color.getHSBColor( 0.11f, 1, 1 ) );
			match_off_graphics.setStroke( new BasicStroke( (float)lcb_boundary_width ) );
			for( lcbI = 0; lcbI < rrpanel.del_lcb_list.length; lcbI++ ){
				LCB delcb = rrpanel.del_lcb_list[ lcbI ];
				if( start_view > delcb.right_end[ seq ] )
					continue;
				if( end_view < delcb.left_end[ seq ] )
					continue;

				int cur_start_pixel = (int)((double)( delcb.left_end[ seq ] - start_view ) / pixel_width);
				double lcb_length = delcb.right_end[ seq ] - delcb.left_end[ seq ];
				int cur_pixel_width = (int) Math.ceil( lcb_length / pixel_width );
					
				java.awt.geom.RoundRectangle2D.Double lcb_rect = new java.awt.geom.RoundRectangle2D.Double( cur_start_pixel + half_pen_width, half_pen_width, cur_pixel_width - lcb_boundary_width, box_height - lcb_boundary_width, arc_size, arc_size );

				match_off_graphics.fill( lcb_rect );
			}
		}
		
		
		long endtime = System.currentTimeMillis();
//		System.out.println( new String( "Took " ) + ( endtime - starttime ) + " milliseconds. " );
}catch(Exception e){
	e.printStackTrace();
}

	
	}
}	// RRSequencePanel


class MatchPopupMenu extends PopupMenu implements ActionListener, ItemListener {
	LinkedList match_list;
	RearrangementPanel rrpanel;
	int sequence;

	int lcb_id;
	int remove_item = Integer.MAX_VALUE;
	int keep_item = Integer.MAX_VALUE;
	static final String remove_string = "Remove this LCB";
	static final String keep_string = "Keep this LCB";
	
	public MatchPopupMenu( RearrangementPanel rrpanel, int sequence ) {
		this.rrpanel = rrpanel;
		this.sequence = sequence;
		match_list = new LinkedList();
		addActionListener( this );
	}

	public void addLCBOptions( int lcb_id) {
//		remove_item = getItemCount();
//		add( new MenuItem( remove_string ) );
		keep_item = getItemCount();
		JCheckBoxMenuItem keep_checkbox = new JCheckBoxMenuItem( keep_string );
		keep_checkbox.setState( false );
		keep_checkbox.addItemListener( this );
//		add( keep_checkbox );
	}

	public void addMatch( Match m ){
		add( new MenuItem( "Align display to " + m.toString() ));
//		add( new MenuItem( new String( "Show annotation between " ) + 
//			m.starts[ sequence ] + new String(" and ") + (m.starts[ sequence ] + m.lengths[ sequence ] - 1) ) );
		match_list.addLast( m );
	}

    public void actionPerformed(ActionEvent e) {
		try{
			String act_command = e.getActionCommand();
			int itemI = 0;
			for(; itemI < getItemCount(); itemI++ ){
				if( getItem(itemI).getActionCommand() == act_command )
					break;
			}
			if( itemI == remove_item ){
//				rrpanel.removeLCB( lcb_id );
			}else if( itemI == keep_item ){
//				rrpanel.keepLCB( lcb_id );
			}else{
				if( itemI > keep_item )
					itemI -= 1;
//					itemI -= 2;
				int matchI = itemI;
	//			int choice = itemI % 2;	// 0 for align sequences, 1 for show annotation
				
				Match selected_match = (Match) match_list.get( matchI );
				rrpanel.alignView( selected_match, sequence );
			}
		}catch( Exception ex ){}
    }
	
	/** this gets called when the "Keep LCB" item is selected */
	public void itemStateChanged(ItemEvent e) {
//		rrpanel.toggleKeepLCB( lcb_id );
	}

}

/**
 * Note: this comparator imposes orderings that are inconsistent with the equals operator.
 */
class MatchStartCompare implements Comparator {
	int seq;

	public MatchStartCompare( int sequence ) {
		int seq = sequence;
	}
	
	public MatchStartCompare( MatchStartCompare m ) {
		int seq = m.seq;
	}

	public int compare( Object o1, Object o2 ) {
		Match a = (Match) o1;
		Match b = (Match) o2;

		long rval = a.starts[ seq ] - b.starts[ seq ];
//		if( rval == 0 )
//			rval = a.lengths[ seq ] - b.lengths[ seq ];
		if( rval < 0 )
			return -1;
		return rval == 0 ? 0 : 1;
	}
	
	public boolean equals( Comparator c ) {
		try{
			return seq == ((MatchStartCompare)c).seq;
		} catch ( Exception e ){
			return false;
		}
	}
}

