package org.gel.mauve;


import java.awt.event.*;
import java.applet.Applet;
import java.net.*;
import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.print.*;
import java.security.*;
import javax.swing.*;
import javax.swing.event.*;

//*< Compares two matches based on generalized offset */
class MatchOffsetCompare implements Comparator {
	static MatchOffsetCompare comp = new MatchOffsetCompare();

	public int compare( Object o1, Object o2 ) {
		Match a = (Match) o1;
		Match b = (Match) o2;
		return (int) (a.offset() - b.offset());
	}
	
	public boolean equals( Comparator c ) {
		try{
			return this == (MatchOffsetCompare) c;
		} catch ( Exception e ){
			return false;
		}
	}
}

//*< Compares two matches based on their multiplicity type */
class MultiplicityTypeCompare implements Comparator {
	static MultiplicityTypeCompare comp = new MultiplicityTypeCompare();

	public int compare( Object o1, Object o2 ) {
		Match a = (Match) o1;
		Match b = (Match) o2;
		return (int) (a.multiplicityType() - b.multiplicityType());
	}
	
	public boolean equals( Comparator c ) {
		try{
			return this == (MultiplicityTypeCompare) c;
		} catch ( Exception e ){
			return false;
		}
	}
}


class LcbLinePanel extends JComponent{
	boolean[] highlighted;	/**< The lcbs that should be highlighted */
	RearrangementPanel rrpanel;
	int menubar_height;
	boolean draw_strikethrough = true;
	
	LcbLinePanel( RearrangementPanel rrpanel ){
		this.rrpanel = rrpanel;
		this.menubar_height = 0;
		highlighted = new boolean[ rrpanel.lcb_list.length ];
		for( int hI = 0; hI < highlighted.length; hI++ ){
			highlighted[ hI ] = true;
		}
	}

	public void paintComponent( Graphics g ) {
		paint( g );
	}
	public void paint( Graphics g1d ) {
try{
		g1d.setColor( Color.black );
		Graphics2D g = (Graphics2D)g1d;
		Rectangle bounds = getBounds();
//		g.setStroke( new BasicStroke( 3f ) );
		for( int seqI = 1; seqI < rrpanel.sequenceCount; seqI++ ){
			Rectangle upper = new Rectangle();
			Rectangle lower = new Rectangle();
			RRSequencePanel upper_p = (RRSequencePanel)rrpanel.sequencePanels.elementAt( seqI - 1 );
			RRSequencePanel lower_p = (RRSequencePanel)rrpanel.sequencePanels.elementAt( seqI );
			int upper_bt = upper_p.boxTop( upper_p.getSize() );
			int upper_bh = upper_p.boxHeight( upper_p.getSize() );
			int lower_bt = lower_p.boxTop( lower_p.getSize() );
			int lower_bh = lower_p.boxHeight( lower_p.getSize() );

			upper_p.getBounds( upper );
			lower_p.getBounds( lower );
			
			for( int lcbI = 0; lcbI < rrpanel.lcb_list.length; lcbI++ ){
				if( !highlighted[ lcbI ] )
					continue;
				if( rrpanel.lcb_list[ lcbI ].color != null )
					g.setColor( rrpanel.lcb_list[ lcbI ].color );
				else
					g1d.setColor( Color.black );
				int upper_midpoint = (upper_p.sequenceCoordinateToPixel( rrpanel.lcb_list[ lcbI ].left_end[ seqI - 1] )
									+ upper_p.sequenceCoordinateToPixel( rrpanel.lcb_list[ lcbI ].right_end[ seqI - 1] ) ) / 2;
				
				int lower_midpoint = (lower_p.sequenceCoordinateToPixel( rrpanel.lcb_list[ lcbI ].left_end[ seqI ] )
									+ lower_p.sequenceCoordinateToPixel( rrpanel.lcb_list[ lcbI ].right_end[ seqI ] ) ) / 2;
				
				
				// draw the connecting bars if at least one of the LCBs is visible
				if( ( upper_midpoint > 0 && upper_midpoint < bounds.width ) ||
					( lower_midpoint > 0 && lower_midpoint < bounds.width ) ){
					// some of this is redundant drawing but who cares
					if( seqI != 1 ){
						if( draw_strikethrough ){
							int y_offset = rrpanel.lcb_list[ lcbI ].reverse[ seqI ] ? 0 : upper_bh / 2;
							g.drawLine( upper_midpoint, upper.y + upper_bt + y_offset, upper_midpoint, upper.y + upper_bt + y_offset + upper_bh / 2);
						}
						g.drawLine( upper_midpoint, upper.y + upper_bt + upper_bh, lower_midpoint, lower.y + lower_bt );
					}else
						g.drawLine( upper_midpoint, upper.y + upper_bt + upper_bh / 2, lower_midpoint, lower.y + lower_bt );
					if( draw_strikethrough ){
						if( seqI + 1 < rrpanel.sequenceCount ){
							int y_offset = rrpanel.lcb_list[ lcbI ].reverse[ seqI ] ? 0 : upper_bh / 2;
							g.drawLine( lower_midpoint, lower.y + lower_bt + y_offset, lower_midpoint, lower.y + lower_bt + y_offset + lower_bh / 2);
						}else if( rrpanel.lcb_list[ lcbI ].reverse[ seqI ] )
							g.drawLine( lower_midpoint, lower.y + lower_bt, lower_midpoint, lower.y + lower_bt + lower_bh / 2 );
					}
				}
			}
		}
}catch(Exception e){
	e.printStackTrace();
}
	}
}

public class RearrangementPanel extends JLayeredPane implements KeyListener, ComponentListener, ActionListener, ChangeListener, java.awt.print.Printable {

	int sequenceCount;		/**< The number of sequences being displayed */
	Vector matchVector;		/**< The current list of all the matches between sequences */
	Match[] original_match_list;	/**< The original full list of all the matches between sequences */
	Vector sequencePanels;	/**< The panels which correspond to each sequence. */
	Vector rulerPanels;		/**< AnimatedRulerPanels threads for each sequence */
	Vector matchColorTable;	/**< Colors to use for different types of matches */
	Vector sequenceLengths;	/**< The length of each sequence */
	Vector sequenceNames;	/**< The name of each sequence */
	LcbLinePanel lcb_line_panel;
	
	LinkedList highBoxes;	/**< A list of RRBoxes which are currently highlighted */
	Match[] matchesByOffset;	/**< A list of the matches sorted by their generalized offset */
	
	LCB[] lcb_list;			/**< The sequence coordinates of currently viewed LCB boundaries */
	LCB[] full_lcb_list;	/**< The sequence coordinates of the complete set of LCB boundaries */
	LCB[] del_lcb_list;		/**< The list of deleted LCBs that should not be shown */

	// Configurable values
	static final int rulerHeight = 20;	/**< Default minimum height for a ruler */
	static final float ruler_ratio = .25f;	/**< Default amount of space for ruler relative to seq panel */
	static final float match_sat = .8f;
	static final float match_bright = .65f;
	static final float lcb_bright = .55f;
	static final float spectrum_gap = .2f; /**< Put a gap in the spectrum around violet to make offsets on each end distinguishable */
	
	static final int CS_SELECT = 0;
	static final int CS_UNIV_ZOOM = 1;
	static final int CS_UNIV_MOVE = 2;
	static final int CS_SEQ_ZOOM = 3;
	static final int CS_SEQ_MOVE = 4;
	
	static final int COLOR_MULTIPLICITY = 0;	/**< Indicates that matches should be colored by their multiplicity */
	static final int COLOR_OFFSET = 1;		/**< Color matches by their generalized offset */
	static final int COLOR_MULTIPLICITY_TYPE = 3;	/**< Color matches by their multiplicity and the sequences they match in */
	static final int COLOR_NORMALIZED_OFFSET = 4;	/**< Color matches by their generalized offset on a normalized color spectrum */
	static final int COLOR_LCB_NUMBER = 5;	/**< Color matches by their LCB number */
	static final int COLOR_NORMALIZED_MULTIPLICITY_TYPE = 6;	/**< Color matches by their multiplicity type on a normalized color spectrum */
	
	boolean lcb_color_computed = false;
	
	public static final Color bg_color = new Color(.85f,.85f,.85f);
	
	long unique_offset_count;
	long lcb_count = 0;
	int cursor_state;
	int selected_sequence;
	int color_type;

	long max_offset;
	long min_offset;

	long longest_range = -1;
	int longest_sequence;
	boolean hand_mode = false;	/**< true when the cursor is in 'hand mode' allowing sequences to be reordered */
	
	JToolBar toolbar;	/**< A toolbar for various alignment manipulation tools, given by the parent frame */
	JLabel weight_label = new JLabel( "LCB weight:" );
	javax.swing.JSlider weight_slider = new javax.swing.JSlider(0, 100, 0);
	JTextField weight_value_text = new JTextField(5);
	
	/** buttons for navigation */
	JButton home_button = new JButton( MauveFrame.home_button_icon );
	JButton left_button = new JButton( MauveFrame.left_button_icon );
	JButton right_button = new JButton( MauveFrame.right_button_icon );
	JButton zoomin_button = new JButton( MauveFrame.zoomin_button_icon );
	JButton zoomout_button = new JButton( MauveFrame.zoomout_button_icon );
	JToggleButton hand_button = new JToggleButton( MauveFrame.hand_button_icon );
	
	/** color scheme selector */
	String[] color_schemes = { "LCB", "Offset", "Normalized offset", 
	"Multiplicity", "Multiplicity type", "Normalized multiplicity type" };
	JComboBox color_selector = new JComboBox( color_schemes );

	Vector lcb_change_points;	/**< Tracks weight values which cause LCBs to drop out */
	
	MauveFrame mauve_frame;	/**< The frame containing this RearrangementPanel */

    public RearrangementPanel ( MauveFrame mauve_frame ) {
	
		addKeyListener( this );
		addComponentListener( this );
		highBoxes = new LinkedList();
		matchColorTable = new Vector();
		selected_sequence = -1;
		color_type = COLOR_OFFSET;
		max_offset = 0;
		min_offset = 0;

		InitZoomCursor();
		sequencePanels = new Vector();
		rulerPanels = new Vector();
		
		this.mauve_frame = mauve_frame;
		this.toolbar = mauve_frame.toolbar;
	}
	
	protected void InitZoomCursor() {
		Dimension D = getToolkit().getBestCursorSize( 16, 16 );
		System.out.println( "Best cursor size: " + D.toString() );
	}

	protected void reorderSequences( int new_order[] ){
		long[] tmp_starts = new long[ sequenceCount ];
		long[] tmp_lengths = new long[ sequenceCount ];
		boolean[] tmp_reverse = new boolean[ sequenceCount ];
		int seqI;

		HashSet reverse_lcbs = new HashSet();
		for( int matchI = 0; matchI < matchVector.size(); matchI++ ){
			Match m = (Match)matchVector.elementAt( matchI );
			for( seqI = 0; seqI < sequenceCount; seqI++ ){
				tmp_starts[ seqI ] = m.starts[ new_order[ seqI ] ];
				tmp_lengths[ seqI ] = m.lengths[ new_order[ seqI ] ];
				tmp_reverse[ seqI ] = m.reverse[ new_order[ seqI ] ];
			}
			System.arraycopy( tmp_starts, 0, m.starts, 0, tmp_starts.length );
			System.arraycopy( tmp_lengths, 0, m.lengths, 0, tmp_starts.length );
			System.arraycopy( tmp_reverse, 0, m.reverse, 0, tmp_starts.length );
			
			for( seqI = 0; seqI < sequenceCount; seqI++ ){
				if( tmp_starts[ seqI ] == 0 )
					continue;
				if( tmp_reverse[ seqI ] == false )
					break;
				// invert the rest of the sequences
				if( lcb_count > 0 )
					reverse_lcbs.add( new Integer( m.lcb ) );
				else
					for( int seqJ = seqI; seqJ < sequenceCount; seqJ++ )
						m.reverse[ seqJ ] = !m.reverse[ seqJ ];
				break;
			}
		}
		// correct the orientation of subset matches within LCBs
		if( lcb_count > 0 ){
			for( int matchI = 0; matchI < matchVector.size(); matchI++ ){
				Match m = (Match)matchVector.elementAt( matchI );
				if( !reverse_lcbs.contains( new Integer( m.lcb ) ) )
					continue;

				for( int seqJ = 0; seqJ < sequenceCount; seqJ++ )
					m.reverse[ seqJ ] = !m.reverse[ seqJ ];
			}
		}
		
		// reorder the panel related data
		Vector tmp_sequencePanels = new Vector();
		Vector tmp_rulerPanels = new Vector();
		Vector tmp_sequenceLengths = new Vector();
		Vector tmp_sequenceNames = new Vector();
		Vector new_panel_vec = new Vector();
		for( seqI = 0; seqI < sequenceCount; seqI++ ){
			tmp_sequencePanels.addElement( sequencePanels.elementAt( new_order[ seqI ] ) );
			tmp_rulerPanels.addElement( rulerPanels.elementAt( new_order[ seqI ] ) );
			tmp_sequenceLengths.addElement( sequenceLengths.elementAt( new_order[ seqI ] ) );
			tmp_sequenceNames.addElement( sequenceNames.elementAt( new_order[ seqI ] ) );
		}
		sequencePanels = tmp_sequencePanels;
		rulerPanels = tmp_rulerPanels;
		sequenceLengths = tmp_sequenceLengths;
		sequenceNames = tmp_sequenceNames;
				
		// set the panel sequence identifiers and reorder the panels themselves
		for( seqI = 0; seqI < sequenceCount; seqI++ ){
			int index;
			index = getIndexOf( (RulerPanel)rulerPanels.elementAt( seqI ) );
			remove( index );
			index = getIndexOf( (RRSequencePanel)sequencePanels.elementAt( seqI ) );
			remove( index );
		}
		validate();

		for( seqI = 0; seqI < sequenceCount; seqI++ ){
			// add elements
			add( (RulerPanel)rulerPanels.elementAt( seqI ), JLayeredPane.DEFAULT_LAYER );
			add( (RRSequencePanel)sequencePanels.elementAt( seqI ), JLayeredPane.DEFAULT_LAYER );
		}
		validate();

		reorderLCBlist( full_lcb_list, new_order );
		reorderLCBlist( lcb_list, new_order );

		for( seqI = 0; seqI < sequenceCount; seqI++ ){
			((RRSequencePanel)sequencePanels.elementAt( seqI )).setSequence( seqI );
		}
		
		if( xmfa != null ){
			try{
				xmfa.reorder( new_order );
				// recalculate sim index since LCB orientation may have changed
				SimilarityIndex[] tmp_sim = new SimilarityIndex[ sequenceCount ];
				for( seqI = 0; seqI < sequenceCount; seqI++ ){
					tmp_sim[ seqI ] = sim[ new_order[ seqI ] ];
					tmp_sim[ seqI ].setSequence( seqI );
					tmp_sim[ seqI ].calculateIndex();
				}
				System.arraycopy( tmp_sim, 0, sim, 0, tmp_sim.length );				
			}catch( Exception e ){
				e.printStackTrace();
			}
		}

		layoutDisplay();
		
	}

	/**
	 * reorders the data in a set of LCBs to a new sequence ordering
	 * called by reorderSequences()
	 */
	public void reorderLCBlist( LCB[] reorder_list, int[] new_order){
	
		long[] tmp_blah = new long[ sequenceCount ];
		int [] int_blah = new int[ sequenceCount ];
		boolean[] tmp_rev = new boolean[ sequenceCount ];
		for( int lcbI = 0; lcbI < reorder_list.length; lcbI++ ){
			int seqI;
			for( seqI = 0; seqI < sequenceCount; seqI++ )
				tmp_blah[ seqI ] = reorder_list[ lcbI ].left_end[ new_order[ seqI ] ];
			System.arraycopy( tmp_blah, 0, reorder_list[ lcbI ].left_end, 0, tmp_blah.length );

			for( seqI = 0; seqI < sequenceCount; seqI++ )
				tmp_blah[ seqI ] = reorder_list[ lcbI ].right_end[ new_order[ seqI ] ];
			System.arraycopy( tmp_blah, 0, reorder_list[ lcbI ].right_end, 0, tmp_blah.length );
			
			for( seqI = 0; seqI < sequenceCount; seqI++ )
				int_blah[ seqI ] = reorder_list[ lcbI ].left_adjacency[ new_order[ seqI ] ];
			System.arraycopy( int_blah, 0, reorder_list[ lcbI ].left_adjacency, 0, int_blah.length );

			for( seqI = 0; seqI < sequenceCount; seqI++ )
				int_blah[ seqI ] = reorder_list[ lcbI ].right_adjacency[ new_order[ seqI ] ];
			System.arraycopy( int_blah, 0, reorder_list[ lcbI ].right_adjacency, 0, int_blah.length );

			for( seqI = 0; seqI < sequenceCount; seqI++ )
				tmp_rev[ seqI ] = reorder_list[ lcbI ].reverse[ new_order[ seqI ] ];
			System.arraycopy( tmp_rev, 0, reorder_list[ lcbI ].reverse, 0, tmp_rev.length );

			// correct the orientation so the first sequence becomes reference orientation
			if( reorder_list[ lcbI ].reverse[ 0 ] )
				for( seqI = 0; seqI < sequenceCount; seqI++ )
					reorder_list[ lcbI ].reverse[ seqI ] = !reorder_list[ lcbI ].reverse[ seqI ];
		}
	}

	/**
	 * Set the match data that this RearrangementPanel will display.
	 * @param names The names of each sequence.
	 * @param lengths The length of each sequence.
	 * @param data_vec A Vector of strings describing matches
	 * @see Match
	 */
	public void SetMatchData( Vector names, Vector lengths, Vector data_vec ) {
		// initialize the color table based on the number of sequences.
		if( data_vec.size() > 0 ){
			sequenceCount = lengths.size();
			sequenceLengths = lengths;
			sequenceNames = names;
		}else
			// no data to display
			return;

		InitColorTable();
		ParseMatches( data_vec );
		InitMatchDisplay();
		colorByOffset();
	}

	protected XMFAAlignment xmfa;
	protected SimilarityIndex[] sim;
	
	public void readXMFAAlignment( RandomAccessFile xmfa_file ) throws IOException, Exception{
	try{
		xmfa = new XMFAAlignment( xmfa_file );
		sim = new SimilarityIndex[ xmfa.seq_count ];
		sequenceLengths = new Vector();
		sequenceNames = new Vector();
		for( int seqI = 0; seqI < xmfa.seq_count; seqI++ ){
			sim[ seqI ] = new SimilarityIndex( seqI, xmfa );
			sequenceLengths.addElement( new Long( xmfa.seq_length[ seqI ] ) );
			// set the sequence name to the comment on the first LCB			
			sequenceNames.addElement( xmfa.getName( seqI ) );
		}
		sequenceCount = xmfa.seq_count;
		matchVector = new Vector();
		matchesByOffset = null;
		
		// init odd match data structures
		original_match_list = new Match[0];

		// copy the LCB list
		full_lcb_list = new LCB[ xmfa.lcb_list.length ];
		System.arraycopy( xmfa.lcb_list, 0, full_lcb_list, 0, xmfa.lcb_list.length );

		del_lcb_list = new LCB[0];

		lcb_color_computed = false;
		lcb_count = full_lcb_list.length;
		InitColorTable();
		InitMatchDisplay();
		colorByLCB();

		if( full_lcb_list.length > 0 )
			initLCBTools();
		
		// disable the color selector since different color modes have
		// no effect when browsing in similarity mode
		color_selector.setEnabled( false );
	}catch( Exception e ){
		e.printStackTrace();
		System.out.println("");
	}

	}
	
	public void ReadRearrangementData( RandomAccessFile match_reader ) throws IOException, Exception
//	public void ReadRearrangementData( BufferedReader match_reader ) throws IOException, Exception
	{
		// the first several lines of match_vector will describe the sequence data
		// the format for version 1 is something like this
		// <file format version>
		// <sequence count>
		// <Mer size>
		// <sequence 1 name>
		// <sequence 1 length>
		// ...
		// <sequence n name>
		// <sequence n length>
		// <match count>
		// matches
		// ...
		String delims = "\t";
		String inputLine;
		inputLine = match_reader.readLine();
		Vector params = Mauve.parseParams( inputLine, delims );
		int format_version;
		try{
			format_version = Integer.parseInt( (String) params.elementAt(1) );
		}catch( Exception e ){
			// probably a NumberFormatException because this is an XMFA file
			match_reader.seek(0);
			readXMFAAlignment( match_reader );
			return;
//			throw new Exception("Rearrangement data file unreadable\n");
		}
		if( format_version != 1 && format_version != 2 && 
			format_version != 3 && format_version != 4 )
			throw new Exception("Rearrangement data format version unknown\n");

		// read sequence count
		inputLine = match_reader.readLine();
		params = Mauve.parseParams( inputLine, delims );
		sequenceCount = Integer.parseInt( (String) params.elementAt(1) );

		// read mer size
//		inputLine = match_reader.readLine();
//		params = Mauve.parseParams( inputLine, delims );
//		int mersize = Integer.parseInt( (String) params.elementAt(1) );

		sequenceNames = new Vector();
		sequenceLengths = new Vector();
		matchVector = new Vector();
		matchesByOffset = null;

		for( int seqI = 0; seqI < sequenceCount; seqI++ ){
			// read file name
			inputLine = match_reader.readLine();
		    StringTokenizer st = new StringTokenizer( inputLine, delims );
		    st.nextToken();
			sequenceNames.addElement( st.nextToken() );
			// read sequence length
			inputLine = match_reader.readLine();
			st = new StringTokenizer( inputLine, delims );
			st.nextToken();
			sequenceLengths.addElement( new Long( st.nextToken() ) );
			long seq_len = ((Long)sequenceLengths.elementAt( 0 )).longValue();
		}
		
		// read match count
		inputLine = match_reader.readLine();
		params = Mauve.parseParams( inputLine, delims );
		int match_count = Integer.parseInt( (String) params.elementAt(1) );
		
		if( format_version == 4 ){
			readMauveAlignment( match_reader );
		}

		while ((inputLine = match_reader.readLine()) != null) {

			Match new_match = new Match( sequenceCount );
			new_match.lcb = 0;
			long m_length = 0;
			int tokenI = 0;
		    StringTokenizer st = new StringTokenizer( inputLine, delims );
	    	int startI = 0;
		    while ( st.hasMoreTokens() ) {
				String s = st.nextToken();
				
				if( tokenI == 0 && ( format_version == 1 || format_version == 3 ))
					m_length = Integer.parseInt( s );
				else if( tokenI == 0 && format_version == 2 ){
					new_match.lcb = Integer.parseInt( s );
					if( new_match.lcb + 1 > lcb_count )
						lcb_count = new_match.lcb + 1;
				}else if( tokenI == 1 && format_version == 2 )
					m_length = Integer.parseInt( s );
				else if( tokenI == sequenceCount + 1 && format_version == 3 ){
					tokenI++;
					continue;
				}else if( tokenI > sequenceCount && format_version == 3 ){
					if( Long.parseLong( s ) != 0 )
						throw new Exception("Match data must have all linked inclusions removed\n");
				}else {
					new_match.lengths[ startI ] = m_length;
					new_match.starts[ startI ] = Long.parseLong( s );
					if( new_match.starts[ startI ] < 0 ){
						new_match.reverse[ startI ] = true;
						new_match.starts[ startI ] *= -1;
//						new_match.starts[ startI ] -= new_match.lengths[ startI ];
					}else if( new_match.starts[ startI ] == Match.NO_MATCH )
						new_match.lengths[ startI ] = 0;
					startI++;
				}
				tokenI++;
		    }
		    long m_offset = 0;
			try{
				m_offset = new_match.offset();
			}catch( Exception e ){
				e.printStackTrace();
			}
			max_offset = m_offset > max_offset ? m_offset : max_offset;
			min_offset = m_offset < min_offset ? m_offset : min_offset;
			matchVector.addElement( new_match );
			if( matchVector.size() % 1000 == 0 )
				System.out.println( "Processed matches: " + matchVector.size() );
		}
		if( full_lcb_list == null )
			full_lcb_list = new LCB[ 0 ];

//		matchVector.addAll( matchlist );
		original_match_list = new Match[ matchVector.size() ];
		for( int matchI = 0; matchI < matchVector.size(); matchI++ ){
			original_match_list[ matchI ] = new Match( (Match)matchVector.elementAt( matchI ) );
		}
		sim = new SimilarityIndex[ sequenceCount ];
		lcb_list = new LCB[ full_lcb_list.length ];
		del_lcb_list = new LCB[0];
		System.arraycopy( full_lcb_list, 0, lcb_list, 0, lcb_list.length );
		InitColorTable();
		InitMatchDisplay();
		if( format_version == 2 || format_version == 4 )
			colorByLCB();
		else
			colorByNormalizedOffset();

		if( full_lcb_list.length > 0 )
			initLCBTools();
	}
	
	public LCB getLCB( int match_start, int match_end ){
		int seqI;
		LCB lcb = new LCB();
		lcb.left_end = new long[ sequenceCount ];
		lcb.right_end = new long[ sequenceCount ];
		for( seqI = 0; seqI < sequenceCount; seqI++ ){
			lcb.left_end[ seqI ] = 0;
			lcb.right_end[ seqI ] = 0;
		}
		if( match_start >= match_end )
			return null;

		// find left end of lcb in each sequence
		int matchI = match_start;
		long count = 0;
		for(; matchI < match_end; matchI++ ){
			Match m = (Match)matchVector.elementAt( matchI );
			for( seqI = 0; seqI < sequenceCount; seqI++ ){
				if( lcb.left_end[ seqI ] == Match.NO_MATCH ){
					if(	m.starts[ seqI ] != Match.NO_MATCH ){
						if( m.reverse[ seqI ] == false )
							lcb.left_end[ seqI ] = m.starts[ seqI ];
						else
							lcb.left_end[ seqI ] = m.starts[ seqI ] + m.lengths[ seqI ];
					}
				}else
					count++;
			}
			if( count == sequenceCount )
				break;
		}
		// if this LCB is only defined in a single sequence then bail out
		count = 0;
		for( seqI = 0; seqI < sequenceCount; seqI++ ){
			if( lcb.left_end[ seqI ] != Match.NO_MATCH )
				count++;
		}
		if( count < 2 )
			return null;
		
		// find end in each sequence
		count = 0;
		for( matchI = match_end; matchI > match_start; matchI-- ){
			Match m = (Match)matchVector.elementAt( matchI - 1 );
			for( seqI = 0; seqI < sequenceCount; seqI++ ){
				if( lcb.right_end[ seqI ] == Match.NO_MATCH ){
					if(	m.starts[ seqI ] != Match.NO_MATCH ){
						if( m.reverse[ seqI ] == false )
							lcb.right_end[ seqI ] = m.starts[ seqI ] + m.lengths[ seqI ];
						else
							lcb.right_end[ seqI ] = m.starts[ seqI ];
					}
				}else
					count++;
			}
			if( count == sequenceCount )
				break;
		}
		Match m = (Match)matchVector.elementAt( match_start );
		for( seqI = 0; seqI < sequenceCount; seqI++ ){
			long tmp;
			if( m.reverse[ seqI ] ){
				tmp = lcb.left_end[ seqI ];
				lcb.left_end[ seqI ] = lcb.right_end[ seqI ];
				lcb.right_end[ seqI ] = tmp;
			}
		}

		// set LCB weight
		lcb.weight = 0;
		for(matchI = match_start; matchI < match_end; matchI++ ){
			m = (Match)matchVector.elementAt( matchI );
			lcb.weight += m.lengths[ 0 ];
		}
		
		// set keep flag to false
		lcb.keep = false;
		
		// set orientation
		lcb.reverse = new boolean[ sequenceCount ];
		System.arraycopy( m.reverse, 0, lcb.reverse, 0, sequenceCount );

		return lcb;
	}

	public void readMauveAlignment( RandomAccessFile match_reader ) throws IOException, Exception{

		String delims = "\t";
		String inputLine;
		int clustal_line = 0;
		int cur_lcb = 0;
		int lcb_start = 0;
		lcb_count = 1;
		Vector lcbVector = new Vector();
		while ((inputLine = match_reader.readLine()) != null) {
			if( clustal_line == sequenceCount + 2 )
				clustal_line = 0;
			if( clustal_line > 0 ){
				clustal_line++;
				continue;
			}
				
			Match new_match = new Match( sequenceCount );
			new_match.lcb = cur_lcb;
			long m_length = 0;
			int tokenI = 0;
		    StringTokenizer st = new StringTokenizer( inputLine, delims );
	    	int startI = 0;
		    while ( st.hasMoreTokens() ) {
				String s = st.nextToken();
				
				if( tokenI == 0 ){
					if( s.equals( "GappedAlignment" ) || s.equals( "ClustalResult" ) ){
						clustal_line = 1;
						new_match = null;
						break;
					}else if( s.startsWith( "Interval" ) ){
						if( matchVector.size() > 0 ){
							LCB new_lcb = getLCB( lcb_start, matchVector.size() );
							lcb_start = matchVector.size();
							if( new_lcb != null ){
								lcbVector.add( new_lcb );
								cur_lcb++;
								lcb_count++;
							}
							new_match = null;
						}
						break;
					}else
						m_length = Integer.parseInt( s );
				} else {
					new_match.lengths[ startI ] = m_length;
					new_match.starts[ startI ] = Long.parseLong( s );
					if( new_match.starts[ startI ] < 0 ){
						new_match.reverse[ startI ] = true;
						new_match.starts[ startI ] *= -1;
//						new_match.starts[ startI ] -= new_match.lengths[ startI ];
					}else if( new_match.starts[ startI ] == Match.NO_MATCH )
						new_match.lengths[ startI ] = 0;
					startI++;
				}
				tokenI++;
		    }
		    if( tokenI != sequenceCount + 1 )
		    	continue;
		    long m_offset = 0;
			try{
				m_offset = new_match.offset();
			}catch( Exception e ){
				e.printStackTrace();
			}
			max_offset = m_offset > max_offset ? m_offset : max_offset;
			min_offset = m_offset < min_offset ? m_offset : min_offset;
			int count = 0;
			for( int seqI = 0; seqI < new_match.lengths.length; seqI++ )
				if( new_match.lengths[ seqI ] != Match.NO_MATCH )
					count++;
			if( count > 1 )
				matchVector.addElement( new_match );
			if( matchVector.size() % 1000 == 0 )
				System.out.println( "Processed matches: " + matchVector.size() );
		}
		LCB new_lcb = getLCB( lcb_start, matchVector.size() );
		if( new_lcb != null )
			lcbVector.add( new_lcb );
		full_lcb_list = new LCB[ lcbVector.size() ];
		full_lcb_list = (LCB[])lcbVector.toArray( full_lcb_list );
	}


	/**
	 * Adds a highlighted RRBox to the list of currently highlighted boxes.  This method does
	 * not highlight the box, it ensures that when another group of boxes are highlighted, this
	 * box is correctly dehighlighted.
	 */
	public void addHighlighted( Match highlighted_box ) {
		ListIterator high_iter = highBoxes.listIterator();
		while( high_iter.hasNext() ){
			if( high_iter.next() == highlighted_box )
				return;
		}
		highBoxes.addLast( highlighted_box );
	}
	
	/**
	 * Unhighlights any boxes which are currently highlighted.
	 */
	public void clearHighlights() {
		ListIterator high_iter = highBoxes.listIterator();
		while( high_iter.hasNext() ){
			Match next_box = (Match)high_iter.next();
			next_box.highlighted = false;
		}
		highBoxes.clear();
	}
	
	protected void ParseMatches( Vector data_vec ) {
		matchVector = new Vector();
		matchesByOffset = null;
		// parse each line of input for matches
		for( int matchI = 0; matchI < data_vec.size(); matchI++ ){
			Vector mvec = Mauve.parseParams( (String) data_vec.elementAt( matchI ) );
			Match new_match = new Match( sequenceCount );

			// parse the size and position of the match
			long m_length = Integer.parseInt( (String)mvec.elementAt( 0 ) );
			for( int seqI = 1; seqI < mvec.size(); seqI++ ){
				new_match.starts[ seqI - 1 ] = Long.parseLong( (String)mvec.elementAt( seqI ) );
				new_match.lengths[ seqI - 1 ] = m_length;
				if( new_match.starts[ seqI - 1 ] < 0 ){
					new_match.reverse[ seqI - 1 ] = true;
					new_match.starts[ seqI - 1 ] *= -1;
				}else if( new_match.starts[ seqI - 1 ] < Match.NO_MATCH )
					new_match.lengths[ seqI - 1 ] = 0;
			}

			long m_offset = new_match.offset();
			max_offset = m_offset > max_offset ? m_offset : max_offset;
			min_offset = m_offset < min_offset ? m_offset : min_offset;

			matchVector.addElement( new_match );
		}
	}

	/**
	 * Colors the matches based on their generalized offsets using the entire color spectrum evenly.
	 * Uses a very simple linear mapping approach for now.  It can be tricked up later.
	 */
	public void colorByNormalizedOffset(){
		color_type = COLOR_NORMALIZED_OFFSET;
		if( matchVector.size() == 0 )
			return;		// no sense in trying to color what's not there!

		// if the array of matches sorted by offset hasn't been created then create it!
		if( matchesByOffset == null ){
			matchesByOffset = new Match[ matchVector.size() ];
			matchesByOffset = (Match[])matchVector.toArray( matchesByOffset );
			Arrays.sort( matchesByOffset, MatchOffsetCompare.comp );
			unique_offset_count = 1;
			long last_offset = matchesByOffset[ 0 ].offset();
			// go thru the sorted list in order counting the number of different offsets
			for( int matchI = 1; matchI < matchesByOffset.length; matchI++ ){
				long cur_offset = matchesByOffset[ matchI ].offset();
				if( cur_offset != last_offset )
					unique_offset_count++;
				last_offset = cur_offset;
			}
		}
		
		// 
		long match_count = matchesByOffset.length;
		double spectrum_bump = spectrum_gap / 2d;
		long cur_offset_count = 0;
		long prev_offset = matchesByOffset[ 0 ].offset();
		// go thru the sorted list in order assigning a new color to each new offset
		for( int matchI = 0; matchI < matchesByOffset.length; matchI++ ){
			long cur_offset = matchesByOffset[ matchI ].offset();
			if( cur_offset != prev_offset )
				cur_offset_count++;
			prev_offset = cur_offset;
			
			// map the generalized offset of this match to a hue
//			Match cur_match = (Match) matchVector.elementAt( matchI );
			Match cur_match = matchesByOffset[ matchI ];
			float hue = (float) ((( 1d - spectrum_gap ) * ((double)cur_offset_count / (double)unique_offset_count)) + spectrum_bump);
			cur_match.color = Color.getHSBColor( hue, match_sat, match_bright);
		}
	}
	
	/**
	 * highlights regions aligned to a particular sequence's coordinate
	 */
	public void highlightRegions( int sequence, long position ){
		long[] iv_col = xmfa.getLCBAndColumn( sequence, position );
		long[] coords = xmfa.getColumnCoordinates( (int)iv_col[0], iv_col[1] );

		for( int seqI = 0; seqI < sequenceCount; seqI++ ){
			RRSequencePanel seqpanel = (RRSequencePanel) sequencePanels.elementAt( seqI );
			seqpanel.clearHighlights();

			if( coords[ seqI ] == Match.NO_MATCH )
				continue;
			seqpanel.highlight( coords[ seqI ] );
		}

		repaintSeqPanels();
	}

	/**
	 * aligns the display to a particular position of a particular sequence
	 */
	public void alignView( int sequence, long position ){
		long[] iv_col = xmfa.getLCBAndColumn( sequence, position );
		long[] coords = xmfa.getColumnCoordinates( (int)iv_col[0], iv_col[1] );
		alignView( coords, sequence );
	}

	/**
	 * Colors the matches based on their generalized offsets.
	 * Uses a very simple linear mapping.  @see colorByNormalizedOffset for a different approach.
	 */
	public void colorByOffset(){
		color_type = COLOR_OFFSET;

		// color by generalized offset (COLOR_OFFSET)
		long offset_range = max_offset - min_offset;
		// bump offset_range up by a designated amount to stay out of the deep violet part of the spectrum
		offset_range += (offset_range * spectrum_gap);
		long offset_bump = (long)(offset_range * (spectrum_gap / 2));
		for( int matchI = 0; matchI < matchVector.size(); matchI++ ){
			// map the generalized offset of this match to a hue
			Match cur_match = (Match) matchVector.elementAt( matchI );
			long m_offset = cur_match.offset() - min_offset + offset_bump;
			float hue = (float)m_offset / (float)offset_range;
			cur_match.color = Color.getHSBColor( hue, match_sat, match_bright);
		}
	}
	
	/**
	 * Colors the matches based on their multiplicity.
	 */
	public void colorByMultiplicity() {
		color_type = COLOR_MULTIPLICITY;
		for( int matchI = 0; matchI < matchVector.size(); matchI++ ){
			Match cur_match = (Match) matchVector.elementAt( matchI );

			// color the match based on its multiplicity.
			int color_count = 0;
			for( int seqI = 0; seqI < cur_match.starts.length; seqI++ ){
				if( cur_match.starts[ seqI ] != Match.NO_MATCH )
					color_count++;
			}
			color_count -= 2;	// assume that all matches have multiplicity >= 2
			cur_match.color = (Color) matchColorTable.elementAt( color_count );

		}
	}
	
	/**
	 * Color matches by their multiplicity and the sequences they match in.
	 * Will not work with more than 62 sequences.
	 */
	public void colorByMultiplicityType() {
		if( sequenceCount > 62 ){
			System.out.println(" Can't color by multiplicity type with more than 62 sequences!!\n");	
			return;
		}

		color_type = COLOR_MULTIPLICITY_TYPE;
		double mult_range = Math.pow( 2, sequenceCount );

		for( int matchI = 0; matchI < matchVector.size(); matchI++ ){
			Match cur_match = (Match) matchVector.elementAt( matchI );

			// color the match based on its multiplicity.
			long color_type = 0;
			for( int seqI = 0; seqI < cur_match.starts.length; seqI++ ){
				color_type <<= 1;
				if( cur_match.starts[ seqI ] != Match.NO_MATCH )
					color_type |= 1;
			}
			double hue = (double) color_type / mult_range;
			cur_match.color = Color.getHSBColor( (float)hue, match_sat, match_bright);
		}
	}

	/**
	 * Color matches by their multiplicity and the sequences they match in.
	 * Will not work with more than 62 sequences.
	 */
	public void colorByNormalizedMultiplicityType() {

		if( sequenceCount > 62 ){
			System.out.println(" Can't color by multiplicity type with more than 62 sequences!!\n");	
			return;
		}

		color_type = COLOR_NORMALIZED_MULTIPLICITY_TYPE;
		if( matchVector.size() == 0 )
			return;		// no sense in trying to color what's not there!
		
		Match[] matchesByMT = null;
		long unique_mt_count = 0;
		if( matchesByMT == null ){
			matchesByMT = new Match[ matchVector.size() ];
			matchesByMT = (Match[])matchVector.toArray( matchesByMT );
			Arrays.sort( matchesByMT, MultiplicityTypeCompare.comp );
			unique_mt_count = 1;
			long last_mt = matchesByMT[ 0 ].multiplicityType();
			// go thru the sorted list in order counting the number of different multiplicity types
			for( int matchI = 1; matchI < matchesByMT.length; matchI++ ){
				long cur_mt = matchesByMT[ matchI ].multiplicityType();
				if( cur_mt != last_mt )
					unique_mt_count++;
				last_mt = cur_mt;
			}
		}
		
		// 
		long match_count = matchesByMT.length;
//		double spectrum_bump = spectrum_gap / 2d;
		long cur_mt_count = 0;
		long prev_mt = matchesByMT[ 0 ].multiplicityType();
		// go thru the sorted list in order assigning a new color to each new offset
		for( int matchI = 0; matchI < matchesByMT.length; matchI++ ){
			long cur_mt = matchesByMT[ matchI ].multiplicityType();
			if( cur_mt != prev_mt )
				cur_mt_count++;
			prev_mt = cur_mt;
			
			// map the generalized offset of this match to a hue
			Match cur_match = (Match) matchesByMT[ matchI ];
//			float hue = (float) ((( 1d - spectrum_gap ) * ((double)cur_mt_count / (double)unique_mt_count)) + spectrum_bump);
			float hue = (float) ((double)cur_mt_count / (double)unique_mt_count);
			cur_match.color = Color.getHSBColor( hue, match_sat, match_bright);
		}
	}

	/**
	 * Color matches by their LCB number
	 */
	public void colorByLCB() {
		// if LCB colors have already been assigned, use those
		if( lcb_color_computed ){
			for( int matchI = 0; matchI < matchVector.size(); matchI++ ){
				Match cur_match = (Match) matchVector.elementAt( matchI );
				if( cur_match.lcb >= 0 )
					cur_match.color = lcb_list[ cur_match.lcb ].match_color;
			}
			return;
		}

		lcb_color_computed = true;
		color_type = COLOR_LCB_NUMBER;
 		double bump_size = 1d / 6d;
 		double color_increment = 1d / (double)lcb_count;
 		// first color the LCBs
 		for( int lcbI = 0; lcbI < full_lcb_list.length; lcbI++ ){
 			double wrap_bump = ( (double)lcbI * bump_size ) / 1d;
 			double color_val = ( (double)lcbI * bump_size ) % 1d;

			// color the match based on its LCB number.
 			double hue = wrap_bump * color_increment + color_val;
			full_lcb_list[ lcbI ].match_color = Color.getHSBColor( (float)hue, match_sat, match_bright);
			full_lcb_list[ lcbI ].color = Color.getHSBColor( (float)hue, match_sat, lcb_bright);
 		}
 		
 		// then color the matches
		for( int matchI = 0; matchI < matchVector.size(); matchI++ ){
			Match cur_match = (Match) matchVector.elementAt( matchI );
			if( cur_match.lcb >= 0 )
				cur_match.color = lcb_list[ cur_match.lcb ].match_color;
		}
	}

	protected void InitColorTable() {
		
		matchColorTable.clear();
		if( sequenceCount == 0 )
			return;	// divide by zero == computer on acid
			
		// modulate the colors through the spectrum for different multiplicities
		// 
		// allow a fixed number of multiplicities for each modulation through the spectrum
		int max_m = 3;	// 5 for testing purposes.
		
		int cycles = (int)Math.ceil( (double)sequenceCount / (double)max_m );

		for( int seqI=0; seqI < sequenceCount; seqI++ ) {
			float hue = ( (float)seqI / (float)max_m ) % 1;
			// add the cycle increment
			hue += ( (1d / (double)max_m)  / (double)cycles) * Math.floor( (double)seqI / (double)max_m );
			matchColorTable.addElement( Color.getHSBColor( hue, match_sat, match_bright ) );
		}
	}
	

	/**
	 * set up a match panel and an animated ruler for each sequence.
	 * also populate the toolbar and set up the LCB line panel
	 */
	protected void InitMatchDisplay() {
		// TODO:  Figure out how the layout managers work...
//		setLayout( new GridLayout( sequenceCount * 2, 1 ));
		for( int seqI = 0; seqI < sequenceCount; seqI++ ){
			if( longest_range < ((Long)sequenceLengths.elementAt( seqI )).longValue() ){
				longest_range = ((Long)sequenceLengths.elementAt( seqI )).longValue();
				longest_sequence = seqI;
			}
		}

		for( int seqI = 0; seqI < sequenceCount; seqI++ ){
			RulerPanel newruler = new RulerPanel( 0, longest_range );
			RRSequencePanel newpanel = new RRSequencePanel( this, matchVector, seqI, (String) sequenceNames.elementAt( seqI ), ((Long)sequenceLengths.elementAt( seqI )).longValue(), sim[ seqI ] );
			newruler.setOpaque( true );
			newpanel.setOpaque( true );
			// add a ruler
			add( newruler, JLayeredPane.DEFAULT_LAYER );
//			newruler.setSize( getSize().width, rulerHeight );
			rulerPanels.add( newruler );
			
			// add a sequence panel
			add( newpanel, JLayeredPane.DEFAULT_LAYER );
			newpanel.setViewableRange( 0l, longest_range );
			newpanel.setVisible( true );
			sequencePanels.add( newpanel );
			
			newpanel.addKeyListener( this );
			newruler.addKeyListener( this );
		}
		
		// set up key bindings
		getInputMap( WHEN_IN_FOCUSED_WINDOW ).put( KeyStroke.getKeyStroke( "ctrl UP" ), "ZoomIn" );
		getInputMap( WHEN_IN_FOCUSED_WINDOW ).put( KeyStroke.getKeyStroke( "ctrl DOWN" ), "ZoomOut" );
		getInputMap( WHEN_IN_FOCUSED_WINDOW ).put( KeyStroke.getKeyStroke( "ctrl LEFT" ), "ScrollLeft" );
		getInputMap( WHEN_IN_FOCUSED_WINDOW ).put( KeyStroke.getKeyStroke( "ctrl RIGHT" ), "ScrollRight" );
		getInputMap( WHEN_IN_FOCUSED_WINDOW ).put( KeyStroke.getKeyStroke( "shift ctrl LEFT" ), "ShiftLeft" );
		getInputMap( WHEN_IN_FOCUSED_WINDOW ).put( KeyStroke.getKeyStroke( "shift ctrl RIGHT" ), "ShiftRight" );

		getInputMap( WHEN_IN_FOCUSED_WINDOW ).put( KeyStroke.getKeyStroke( "r" ), "ToggleStrikethrough" );
		getInputMap( WHEN_IN_FOCUSED_WINDOW ).put( KeyStroke.getKeyStroke( "shift l" ), "ToggleLCBlines" );
		getInputMap( WHEN_IN_FOCUSED_WINDOW ).put( KeyStroke.getKeyStroke( "h" ), "ToggleHandMode" );

		getActionMap().put("ZoomIn", new GenericAction( this, "ZoomIn" ) );
		getActionMap().put("ZoomOut", new GenericAction( this, "ZoomOut" ) );
		getActionMap().put("ScrollLeft", new GenericAction( this, "ScrollLeft" ) );
		getActionMap().put("ScrollRight", new GenericAction( this, "ScrollRight" ) );
		getActionMap().put("ShiftLeft", new GenericAction( this, "ShiftLeft" ) );
		getActionMap().put("ShiftRight", new GenericAction( this, "ShiftRight" ) );

		getActionMap().put("ToggleStrikethrough", new GenericAction( this, "ToggleStrikethrough" ) );
		getActionMap().put("ToggleLCBlines", new GenericAction( this, "ToggleLCBlines" ) );
		getActionMap().put("ToggleHandMode", new GenericAction( this, "ToggleHandMode" ) );


		// populate the toolbar
		toolbar.removeAll();
//		toolbar.add( Box.createHorizontalGlue() );
		home_button.setToolTipText( "Reset display" );
		left_button.setToolTipText( "Shift display left (Ctrl+Left)" );
		right_button.setToolTipText( "Shift display right (Ctrl+Right)" );
		zoomin_button.setToolTipText( "Zoom in (Ctrl+Up)" );
		zoomout_button.setToolTipText( "Zoom out (Ctrl+Down)" );
		hand_button.setToolTipText( "Toggle sequence reordering mode." );
		
		hand_button.setPressedIcon( MauveFrame.dark_hand_button_icon );
		hand_button.setSelectedIcon( MauveFrame.dark_hand_button_icon );
		
		home_button.addActionListener( this );
		left_button.addActionListener( this );
		right_button.addActionListener( this );
		zoomin_button.addActionListener( this );
		zoomout_button.addActionListener( this );
		hand_button.addActionListener( this );

//		toolbar.setRollover( true );
		toolbar.add( home_button );
		toolbar.add( left_button );
		toolbar.add( right_button );
		toolbar.add( zoomin_button );
		toolbar.add( zoomout_button );
		toolbar.add( hand_button );

// add color scheme selection combo box
		color_selector.setSelectedIndex( 1 );
		color_selector.addActionListener( this );
		toolbar.add( color_selector );
		
		Dimension minSize = new Dimension(5, 3);
		Dimension prefSize = new Dimension(5, 3);
		Dimension maxSize = new Dimension(Short.MAX_VALUE, 3);
		toolbar.add(new Box.Filler(minSize, prefSize, maxSize));

	}
	
	protected void initLCBTools() {

		Dimension prefSize = new Dimension(5, 3);
		LCBlist.computeLCBAdjacencies( full_lcb_list );

		// find minimum and maximum weights
		long lcb_minimum_weight = -1;
		long lcb_max_weight = 0;
		for( int lcbI = 0; lcbI < full_lcb_list.length; lcbI++ ){
			if( full_lcb_list[ lcbI ].weight < lcb_minimum_weight || lcb_minimum_weight == -1 )
				lcb_minimum_weight = full_lcb_list[ lcbI ].weight;
		}

		lcb_list = new LCB[ full_lcb_list.length ];
		LCB[] tmp_lcb_list = new LCB[ full_lcb_list.length ];
		for( int lcbI = 0; lcbI < full_lcb_list.length; lcbI++ ){
			lcb_list[ lcbI ] = new LCB( full_lcb_list[ lcbI ] );
			tmp_lcb_list[ lcbI ] = new LCB( full_lcb_list[ lcbI ] );
		}
		lcb_change_points = new Vector();
		LCBlist.greedyBreakpointElimination( 0x7fffffff, tmp_lcb_list, lcb_change_points );
		LCBlist.filterLCBs( tmp_lcb_list, null, null );
		
		lcb_max_weight = tmp_lcb_list[ 0 ].weight;

//
// add interface components
//
		// add the line panel to the top pane
		lcb_line_panel = new LcbLinePanel( this );
		add( lcb_line_panel, JLayeredPane.PALETTE_LAYER );
		lcb_line_panel.setVisible( true );

		// add the weight slider components
		toolbar.add( weight_label );

		weight_slider.setMinimum( 0 );
		weight_slider.setMaximum( lcb_change_points.size() - 1 );
		weight_slider.setMinorTickSpacing( 1 );
		weight_slider.setMajorTickSpacing( (int)(lcb_change_points.size() / 10) );
		weight_slider.setToolTipText( "Set the minimum weight for Locally Collinear Blocks" );
		weight_slider.setPaintTicks( true );
		weight_slider.setPaintLabels( false );
		weight_slider.setSnapToTicks( false );
		prefSize = weight_slider.getPreferredSize();
		prefSize.setSize( 250, prefSize.getHeight() );
		weight_slider.setPreferredSize( prefSize );
		weight_slider.setMaximumSize( prefSize );
		toolbar.add( weight_slider );
		weight_slider.addChangeListener( this );

		prefSize = weight_value_text.getPreferredSize();
		prefSize.setSize( 40, prefSize.getHeight() );
		weight_value_text.setPreferredSize( prefSize );
		weight_value_text.setMaximumSize( prefSize );
		weight_value_text.setHorizontalAlignment( JTextField.RIGHT );
		weight_value_text.setText( lcb_change_points.elementAt(0).toString() );
		toolbar.add( weight_value_text );
		
		// tell the user how many LCBs there are
		String status_text = "There are " + lcb_list.length + " LCBs with minimum weight " + lcb_minimum_weight;
		mauve_frame.status_bar.setHint( status_text );

		// set the appropriate LCB color scheme
		color_selector.setSelectedIndex( 0 );
	}

	protected void layoutDisplay() {
		Dimension dim = getSize();
//		System.out.println( dim.toString() );
		for( int seqI = 0; seqI < sequenceCount; seqI++ ){
			RulerPanel ruler = (RulerPanel) rulerPanels.elementAt( seqI );
			RRSequencePanel seqpanel = (RRSequencePanel) sequencePanels.elementAt( seqI );
			
			double seq_area = dim.height / sequenceCount;
			int r_height = ruler_ratio * seq_area > rulerHeight ? (int)(ruler_ratio * seq_area) : rulerHeight;
			
			ruler.setBounds( 0, (seqI * dim.height) / sequenceCount, dim.width, r_height );
//			ruler.repaint();
			seqpanel.setBounds( 0, (seqI * dim.height) / sequenceCount + r_height, dim.width, (dim.height / sequenceCount) - r_height );
			seqpanel.formatBoxes();
//			seqpanel.repaint(0, 0, dim.width, (dim.height / sequenceCount) - rulerHeight);
		}
		if( lcb_line_panel != null )
			lcb_line_panel.setBounds( 0, 0, dim.width, dim.height );
	}

	public void componentResized(ComponentEvent e){
		layoutDisplay();
	}
	/**< Not implemented */
	public void componentHidden(ComponentEvent e){}
	/**< Not implemented */
	public void componentMoved(ComponentEvent e){}
	/**< Not implemented */
	public void componentShown(ComponentEvent e){}

	/**
	 * zoom_percent 0 will zoom all the way out. 100 does nothing.  values between 1 and 99 zoom out, values greater than 100 zoom in
	 */
	protected void zoom( int zoom_percent, int move_percent ){
		for( int seqI = 0; seqI < sequenceCount; seqI++ ){
			RulerPanel ruler = (RulerPanel) rulerPanels.elementAt( seqI );
			RRSequencePanel seqpanel = (RRSequencePanel) sequencePanels.elementAt( seqI );

			// get the current dimensions
			long[] view_array = new long[2];
			seqpanel.getViewableRange( view_array );
			
			// move before zooming so that we end up on the right coordinates!
			if( move_percent != 0 ){
				// move in some direction
				long cur_move = (long)((( view_array[1] ) * ( move_percent / 100d )));
				// calculate the new viewable coordinates.
				long new_view_start = view_array[0] + cur_move;
				
				// don't let the sequence get completely out of range
				if( new_view_start + view_array[ 1 ] < 1 )
					new_view_start = 1 - view_array[ 1 ];
				else if( new_view_start > ((Long)sequenceLengths.elementAt( seqI )).longValue() )
					new_view_start = ((Long)sequenceLengths.elementAt( seqI )).longValue();

				ruler.setViewableRange( new_view_start, view_array[1] );
				seqpanel.setViewableRange( new_view_start, view_array[1] );
			}

			seqpanel.getViewableRange( view_array );

			if( zoom_percent == 0 ){
				// zoom all the way out.
				ruler.setViewableRange( 1l, longest_range );
				seqpanel.setViewableRange( 1l, longest_range );
			}else if( zoom_percent != 100 ){
				// calculate the new viewable coordinates.
				long new_view_length = (long)((100d * view_array[1] ) / zoom_percent);
				// don't let the view length get too big or small
				if( new_view_length < 5 )
					new_view_length = 5;
				if( new_view_length > longest_range )
					new_view_length = longest_range;

				long new_view_start = view_array[0] + ( ( view_array[1] - new_view_length ) / 2 );

				// don't let the sequence get completely out of range
				if( new_view_start + new_view_length < 1 )
					new_view_start = 1 - view_array[ 1 ];
				else if( new_view_start > ((Long)sequenceLengths.elementAt( seqI )).longValue() )
					new_view_start = ((Long)sequenceLengths.elementAt( seqI )).longValue();

				ruler.setViewableRange( new_view_start, new_view_length );
				seqpanel.setViewableRange( new_view_start, new_view_length );
			}
		}
		if( move_percent != 0 || zoom_percent != 100 )
			repaintSeqPanels();
	}

	/**
	 * Align the display to a particular set of coordinates in each sequence, using
	 * one sequence to set the viewable width
	 */
	public void alignView( long[] align_coords, int sequence ) {
		// get the current dimensions
		long[] view_array = new long[2];
		RRSequencePanel seqpanel = (RRSequencePanel) sequencePanels.elementAt( sequence );
		seqpanel.getViewableRange( view_array );

		// adjust everything to the same zoom level as the selected sequence
		long start_offset = align_coords[sequence] - view_array[0];
		for( int seqI = 0; seqI < sequenceCount; seqI++ ){
			if( seqI == sequence )
				continue;
			if( align_coords[ seqI ] == Match.NO_MATCH )
				continue;

			RulerPanel ruler = (RulerPanel) rulerPanels.elementAt( seqI );
			seqpanel = (RRSequencePanel) sequencePanels.elementAt( seqI );
			long new_view_start = align_coords[ seqI ] - start_offset;
			ruler.setViewableRange( new_view_start , view_array[1] );
			seqpanel.setViewableRange( new_view_start , view_array[1] );
		}

		repaintSeqPanels();
	}

	public void alignView( Match align_match, int sequence ) {
		alignView( align_match.starts, sequence );
	}
	
	/**
	 * rescales all sequences to the same scale as the longest genome, shifting
	 * the display to maintain center points.
	 */ 
	public void lockScales() {
		long[] view_array = new long[2];

		RRSequencePanel seqpanel = (RRSequencePanel) sequencePanels.elementAt( longest_sequence );
		seqpanel.getViewableRange( view_array );
		long cur_long_range = view_array[ 1 ];

		// rescale everything to the scale of the largest sequence range
		for( int seqI = 0; seqI < sequenceCount; seqI++ ){
			seqpanel = (RRSequencePanel) sequencePanels.elementAt( seqI );
			RulerPanel ruler = (RulerPanel) rulerPanels.elementAt( seqI );
			seqpanel.getViewableRange( view_array );
			long range_diff = ( view_array[ 1 ] - cur_long_range ) / 2;
			seqpanel.setViewableRange( view_array[0] + range_diff, cur_long_range );
			ruler.setViewableRange( view_array[0] + range_diff, cur_long_range );
		}
		repaintSeqPanels();
	}

	/**
	 * This method is a quick hack.  don't use it!
	 */
	public void repaintSeqPanels(){
		for( int seqI = 0; seqI < sequenceCount; seqI++ ){
			RRSequencePanel seqpanel = (RRSequencePanel) sequencePanels.elementAt( seqI );
			seqpanel.formatBoxes();
			seqpanel.repaint();
			RulerPanel rpanel = (RulerPanel) rulerPanels.elementAt( seqI );
			rpanel.repaint();
		}
		lcb_line_panel.repaint();
	}
	
	/**
	 * Renders this sequence panel to the specified graphics context for printing.
	 * Assumes there is only one page in the document.
	 */
	public int print(Graphics graphics, PageFormat pageFormat, int pageIndex ) throws PrinterException {
		if( pageIndex > 0 )
			return NO_SUCH_PAGE;
		
		Graphics2D g2d = (Graphics2D)graphics;
//		g2d.transform( new java.awt.geom.AffineTransform( pageFormat.getMatrix() ) );
		
		int ruler_percent = 10;	/**< Rulers will occupy this percentage of the display or rulerHeight, whichever is greater */
		int dpi = 600;
		Rectangle draw_box = new Rectangle( (int)pageFormat.getImageableX(), (int)pageFormat.getImageableY(), (int)( (pageFormat.getImageableWidth() / 72) * dpi ), (int)( (pageFormat.getImageableHeight() / 72) * dpi ) );
		
		Rectangle print_box = graphics.getClipBounds();
		int print_seq_height = print_box.height / sequenceCount;
		int print_r_height = (int) (((double) ruler_percent / 100d) * print_seq_height);

		int seq_height = draw_box.height / sequenceCount;
		int r_height = (int) (((double) ruler_percent / 100d) * seq_height);
		r_height = r_height > rulerHeight ? r_height : rulerHeight;
		
		// construct an Image to draw the rulers into.
		Dimension ruler_d;
		Image ruler_i;
		Graphics2D ruler_g;
	    ruler_d = new Dimension( draw_box.width, r_height );
	    ruler_i = createImage( ruler_d.width, ruler_d.height );
	    ruler_g = (Graphics2D)ruler_i.getGraphics();

		// construct an Image to draw the seqpanels into.
		Dimension seqpanel_d;
		Image seqpanel_i;
		Graphics2D seqpanel_g;
	    seqpanel_d = new Dimension( draw_box.width, seq_height - r_height );
	    seqpanel_i = createImage( seqpanel_d.width, seqpanel_d.height );
	    seqpanel_g = (Graphics2D)seqpanel_i.getGraphics();
		
		long[] view_array = new long[2];
		for( int seqI = 0; seqI < sequenceCount; seqI++ ){
			RRSequencePanel seqpanel = (RRSequencePanel) sequencePanels.elementAt( seqI );
			// get the current dimensions
			seqpanel.getViewableRange( view_array );

			// paint the ruler panel into the graphics object
			RulerPanel ruler = new RulerPanel( (int) view_array[0], (int) (view_array[0] + view_array[1] - 1) );
			ruler.setSize( ruler_d );
			ruler.paint( ruler_g );
			g2d.drawImage( ruler_i, draw_box.x, draw_box.y + seqI * seq_height, null );
			ruler_g.clearRect(0, 0, ruler_d.width, ruler_d.height );

			// paint the sequence panel into the graphics object
//			seqpanel.paintRange( seqpanel_g, seqpanel.getSize(), view_array[0], view_array[1], false );
			seqpanel.paintRange( seqpanel_g, seqpanel_d, view_array[0], view_array[1], false );
			g2d.drawImage( seqpanel_i, print_box.x, print_box.y + seqI * print_seq_height + print_r_height, print_box.x + print_box.width, print_box.y + (seqI + 1 ) * print_seq_height + print_r_height, 0, 0, seqpanel_d.width, seqpanel_d.height, null );
			seqpanel_g.clearRect(0, 0, seqpanel_d.width, seqpanel_d.height );
			
		}

		return PAGE_EXISTS;
	}
	
    public void keyPressed( KeyEvent e ) {
    	switch( e.getKeyCode() ){
			case KeyEvent.VK_Q:
				for( int seqI = 0; seqI < sequenceCount; seqI++ ){
					boolean opposite = !((RRSequencePanel)sequencePanels.elementAt( seqI )).getDrawLCBbounds();
					((RRSequencePanel)sequencePanels.elementAt( seqI )).setDrawLCBbounds( opposite );
				}
				break;
			case KeyEvent.VK_W:
				for( int seqI = 0; seqI < sequenceCount; seqI++ ){
					boolean opposite = !((RRSequencePanel)sequencePanels.elementAt( seqI )).getFillLCBboxes();
					((RRSequencePanel)sequencePanels.elementAt( seqI )).setFillLCBboxes( opposite );
				}
				break;
			case KeyEvent.VK_E:
				for( int seqI = 0; seqI < sequenceCount; seqI++ ){
					boolean opposite = !((RRSequencePanel)sequencePanels.elementAt( seqI )).getDrawMatches();
					((RRSequencePanel)sequencePanels.elementAt( seqI )).setDrawMatches( opposite );
				}
				break;

    		default:
    			break;
    	}
    }

    public void keyReleased( KeyEvent e ) {
    }

    public void keyTyped( KeyEvent e ) {

    }
	
	// CHANGE LISTENER METHODS
	
	public void stateChanged( ChangeEvent e )
	{
		try{
		if( e.getSource() == weight_slider ){
			// compute the updated list of LCBs
			int w_int = weight_slider.getValue();
			int min_weight = ((Integer)lcb_change_points.elementAt( w_int )).intValue();
			weight_value_text.setText( lcb_change_points.elementAt( w_int ).toString() );
			
			lcb_list = new LCB[ full_lcb_list.length ];
			for( int lcbI = 0; lcbI < full_lcb_list.length; lcbI++ )
				lcb_list[ lcbI ] = new LCB( full_lcb_list[ lcbI ] );
			
			LCBlist.greedyBreakpointElimination( min_weight, lcb_list, null );
			
			// toggle the LCBs to and from solid cartoons when the slider is dragged
			if( weight_slider.getValueIsAdjusting() ){
				for( int seqI = 0; seqI < sequenceCount; seqI++ ){
					((RRSequencePanel)sequencePanels.elementAt( seqI )).setFillLCBboxes( true );
					((RRSequencePanel)sequencePanels.elementAt( seqI )).setDrawMatches( false );
					// filter the LCB list appropriately
				}
				lcb_list = LCBlist.filterLCBs( lcb_list, null, null );
				String status_text = "There are " + lcb_list.length + " LCBs with minimum weight " + min_weight;
				mauve_frame.status_bar.setHint( status_text );
			}else{
				// done adjusting the value
				for( int seqI = 0; seqI < sequenceCount; seqI++ ){
					((RRSequencePanel)sequencePanels.elementAt( seqI )).setFillLCBboxes( false );
					((RRSequencePanel)sequencePanels.elementAt( seqI )).setDrawMatches( true );

				}
				// filter the LCB and Match lists appropriately
				for( int matchI = 0; matchI < original_match_list.length; matchI++ ){
					((Match)matchVector.elementAt( matchI )).lcb = original_match_list[ matchI ].lcb;
					((Match)matchVector.elementAt( matchI )).color = original_match_list[ matchI ].color;
				}
				Vector del_vec = new Vector();
				lcb_list = LCBlist.filterLCBs( lcb_list, matchVector, del_vec );
				del_lcb_list = (LCB[])del_vec.elementAt(0);
//				mauve_frame.status_bar.setHint( "" );
			}

			repaintSeqPanels();
		}
		}catch(Exception ae){
			ae.printStackTrace();
		}		
//		public void toggleKeepFlag( int lcb_id ){
//		}
	}
	
	// ACTION LISTENER METHODS
	
	public void actionPerformed(ActionEvent e) {			
		if( e.getSource() == color_selector ){
			if( color_selector.getSelectedItem().equals( "LCB" ) ){
				if( color_type != COLOR_LCB_NUMBER )
					colorByLCB();
			}else if( color_selector.getSelectedItem().equals( "Offset" ) ){
				if( color_type != COLOR_OFFSET )
					colorByOffset();
			}else if( color_selector.getSelectedItem().equals( "Normalized offset" ) ){
				if( color_type != COLOR_NORMALIZED_OFFSET )
					colorByNormalizedOffset();
			}else if( color_selector.getSelectedItem().equals( "Multiplicity" ) ){
				if( color_type != COLOR_MULTIPLICITY )
					colorByMultiplicity();
			}else if( color_selector.getSelectedItem().equals( "Multiplicity type" ) ){
				if( color_type != COLOR_MULTIPLICITY_TYPE )
					colorByMultiplicityType();
			}else if( color_selector.getSelectedItem().equals( "Normalized multiplicity type" ) ){
				if( color_type != COLOR_NORMALIZED_MULTIPLICITY_TYPE )
					colorByNormalizedMultiplicityType();
			}
			repaintSeqPanels();
		}else if( e.getSource() == home_button ){
			zoom( 0, Integer.MIN_VALUE );
//			zoom( 100, 100 );
		}else if( e.getSource() == left_button || 
				e.getActionCommand().equals( "ShiftLeft" ) || 
				e.getActionCommand().equals( "ScrollLeft" ) ){

			if( (e.getModifiers() & InputEvent.SHIFT_MASK) != 0 )
				// big jump left
    			zoom( 100, -50 );
    		else
    			// scroll left
    			zoom( 100, -10 );
		}else if( e.getSource() == right_button ||
				e.getActionCommand().equals( "ShiftRight" ) || 
				e.getActionCommand().equals( "ScrollRight" ) ){
			if( (e.getModifiers() & InputEvent.SHIFT_MASK) != 0 )
				// big jump right
    			zoom( 100, 50 );
    		else
    			// scroll right
    			zoom( 100, 10 );
		}else if( e.getSource() == zoomin_button ||
				e.getActionCommand().equals( "ZoomIn" ) ){
			zoom( 200, 0 );
		}else if( e.getSource() == zoomout_button ||
				e.getActionCommand().equals( "ZoomOut" ) ){
			zoom( 50, 0 );		
		}else if( e.getSource() == hand_button ||
				e.getActionCommand().equals( "ToggleHandMode" ) ){
			if( hand_mode ){
				setCursor( new Cursor( Cursor.DEFAULT_CURSOR ) );
				// add mouse listeners 
				for( int seqI = 0; seqI < sequenceCount; seqI++ ){
					RRSequencePanel rrsp = (RRSequencePanel)sequencePanels.elementAt( seqI );
					rrsp.addMouseListener( rrsp );
					rrsp.addMouseMotionListener( rrsp );
				}
			}else{
				if( MauveFrame.hand_cursor != null )
					setCursor( MauveFrame.hand_cursor );
				else
					setCursor( new Cursor( Cursor.HAND_CURSOR ) );
				for( int seqI = 0; seqI < sequenceCount; seqI++ ){
					RRSequencePanel rrsp = (RRSequencePanel)sequencePanels.elementAt( seqI );
					rrsp.removeMouseListener( rrsp );
					rrsp.removeMouseMotionListener( rrsp );
				}
			}
			hand_mode = !hand_mode;
		}else if( e.getActionCommand().equals( "ToggleLCBlines" ) ){
			lcb_line_panel.setVisible( !lcb_line_panel.isVisible() );
		}else if( e.getActionCommand().equals( "ToggleStrikethrough" ) ){
			lcb_line_panel.draw_strikethrough = !lcb_line_panel.draw_strikethrough;
			lcb_line_panel.repaint();
		}

	}

}

class GenericAction extends AbstractAction {

	GenericAction( ActionListener al, String command ){
		super( command );
		this.al = al;
		putValue( ACTION_COMMAND_KEY, command );
	}

	ActionListener al;
	public void actionPerformed(ActionEvent e){
		al.actionPerformed( e );
	}
}

class PrivilegedPrinting implements PrivilegedAction {
	Printable print_object;
	PageFormat format;

	public PrivilegedPrinting( PageFormat format, Printable print_object ){
		this.print_object = print_object;
		this.format = format;
	}

	public Object run() {
		// privileged code goes here, for example:
		PrinterJob pjob = PrinterJob.getPrinterJob();
		if( !pjob.printDialog())
			return null;
		pjob.setPrintable( print_object, format);
		try{
			pjob.print();
		}catch( PrinterException pe ){ System.out.println(pe); }
		return null;
	}
}
